#ifndef __DB_ANALYZE_H
#define __DB_ANALYZE_H
/*-------------------------------------------------------------------------
 * Copyright (c) 2004 Kenneth W. Sodemann (stuffle@mac.com)
 *-------------------------------------------------------------------------
 * db_analyze
 *
 * Synopsis:
 *    
 * $Id: db_analyze.h,v 1.5 2004/06/29 18:45:41 stuffle Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330 
 * Boston, MA  02111-1307  USA
 *-------------------------------------------------------------------------
 */

#include <glib.h>
#include <libpq-fe.h>


/*-------------------------------------------------------------------------
 * get_db_version
 *
 * Determine the current version of a PRepS schema.  If the schema
 * does not have a version table with a numeric "schema_ver" column,
 * a version of zero will be returned.
 *
 * Inputs:
 *    conn : database connection
 *-------------------------------------------------------------------------
 */
gint get_db_version (PGconn *conn);

/*-------------------------------------------------------------------------
 * db_has_prereq_objets
 *
 * Analyze the database, and determine if at least the minimum objects
 * exist.
 *
 * Inputs:
 *    conn : database connection
 *-------------------------------------------------------------------------
 */
gboolean db_has_prereq_objects (PGconn *conn);

/*-------------------------------------------------------------------------
 * table_exists
 *
 * Determine if a given table exists.
 *
 * Inputs:
 *    conn  : database connection
 *    table : the name of the table
 *-------------------------------------------------------------------------
 */
gboolean table_exists (PGconn *conn, const gchar *table);

/*-------------------------------------------------------------------------
 * table_pk_exists
 *
 * Determine if the given table's primary key exists.
 *
 * Inputs:
 *    conn  : database connection
 *    table : the name of the table
 *-------------------------------------------------------------------------
 */
gboolean table_pk_exists (PGconn *conn, const gchar *table);

/*-------------------------------------------------------------------------
 * col_exists
 *
 * Determine if a given column exists in a table.
 *
 * Inputs:
 *    conn  : database connection
 *    table : the name of the table
 *    col   : the name of the column
 *-------------------------------------------------------------------------
 */
gboolean col_exists (PGconn *conn, const gchar *table, const gchar *col);

/*-------------------------------------------------------------------------
 * lang_exists
 *
 * Determine if a given language exists in the database.
 *
 * Inputs:
 *    conn : database connection
 *    lang : the name of the language
 *-------------------------------------------------------------------------
 */
gboolean lang_exists (PGconn *conn, const gchar *lang);

/*-------------------------------------------------------------------------
 * index_exists
 *
 * Determine if a given index exists in the database.
 *
 * Inputs:
 *    conn : database connection
 *    lang : the name of the language
 *-------------------------------------------------------------------------
 */
gboolean index_exists (PGconn *conn, const gchar *lang);
#endif
