#ifndef __DB_DDL_H
#define __DB_DDL_H
/*-------------------------------------------------------------------------
 * Copyright (c) 2004 Kenneth W. Sodemann (stuffle@mac.com)
 *-------------------------------------------------------------------------
 * db_ddl
 *
 * Synopsis:
 *   Routines used to create the DDL strings u8sed by the other modules
 *   in this package.
 *    
 * $Id: db_ddl.h,v 1.4 2004/06/29 18:45:41 stuffle Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330 
 * Boston, MA  02111-1307  USA
 *-------------------------------------------------------------------------
 */
#include <glib.h>
#include <libpq-fe.h>

#include "db_schema.h"

#define CR8SEQ "create sequence %s start %lu increment %lu \
                maxvalue %lu minvalue %lu cache %lu"

/*-------------------------------------------------------------------------
 * create_table_str
 *
 * Build a create table DDL string.
 *
 * Inputs:
 *    buff    : the GString buffer used to hold the DDL.
 *    idx     : index into the table list
 *    version : database schema version
 *
 *-------------------------------------------------------------------------
 */
void create_table_str (GString *buff, gint idx, gint version);

/*-------------------------------------------------------------------------
 * add_col_str
 *
 * Build an "alter table add column" DDL string.
 *
 * Inputs:
 *    buff   : the GString buffer used to hold the DDL.
 *    t_idx  : index into the table list
 *    c_idx  : index into the table's column list
 *
 *-------------------------------------------------------------------------
 */
void add_col_str (GString *buff, gint t_idx, gint c_idx);


/*-------------------------------------------------------------------------
 * col_not_null_str
 *
 * Build an "alter table alter column set not NULL" DDL string.
 *
 * Inputs:
 *    buff   : the GString buffer used to hold the DDL.
 *    t_idx  : index into the table list
 *    c_idx  : index into the table's column list
 *
 *-------------------------------------------------------------------------
 */
void col_not_null_str (GString *buff, gint t_idx, gint c_idx);

/*-------------------------------------------------------------------------
 * add_primary_key_str
 *
 * Build a alter table add primary key DDL string.
 *
 * Inputs:
 *    buff    : the GString buffer used to hold the DDL.
 *    idx     : index into the table list
 *    version : database schema version
 *
 *-------------------------------------------------------------------------
 */
void add_primary_key_str (GString *buff, gint idx, gint version);

/*-------------------------------------------------------------------------
 * create_unique_index_str
 *
 * Build a create unique index string.
 *
 * Inputs:
 *    buff    : the GString buffer used to hold the DDL.
 *    idx     : index into the unique index list
 *
 *-------------------------------------------------------------------------
 */
void create_unique_index_str (GString *buff, UniqueIndexData *idx);

/*-------------------------------------------------------------------------
 * fix_grants
 *
 * Fix the grants on an object.
 *
 * Inputs:
 *    conn : the database connection
 *    obj  : the name of the object to fix the grants for
 *
 * Returns:
 *    A pointer to the results of the grant fixing.  The caller needs
 *    to PQclear() the result.
 *
 *-------------------------------------------------------------------------
 */
PGresult *fix_grants (PGconn *conn, const gchar *obj);
#endif
