#ifndef __DB_UTIL_H
#define __DB_UTIL_H
/*-------------------------------------------------------------------------
 * Copyright (c) 2003 Kenneth W. Sodemann (stuffle@mac.com)
 *-------------------------------------------------------------------------
 * db_login
 *
 * Synopsis:
 *    Utility functions used for database access.
 *    
 * $Id: db_login.h,v 1.5 2003/11/26 03:11:54 stuffle Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330 
 * Boston, MA  02111-1307  USA
 *-------------------------------------------------------------------------
 */
#include <libpq-fe.h>
#include <glib.h>

/*
 * Simple transaction handlers
 */
#define START_TRANSACTION(conn)  PQclear (PQexec ((conn), "BEGIN"))
#define COMMIT(conn)             PQclear (PQexec ((conn), "COMMIT"))
#define ROLLBACK(conn)           PQclear (PQexec ((conn), "ROLLBACK"))

/*
 * DbLoginData
 *
 * This structure contains the common data that may be needed to
 * log in to a database.
 */
typedef struct DbLoginData {
   GString  *username;
   GString  *password;
   GString  *db_name;
   GString  *descr;
   GString  *host;
   GString  *port;
   PGconn   *conn;
} DbLoginData;

/*-------------------------------------------------------------------------
 * set_db_XXXX
 *
 * Set the given field to the given value.  Allocate space if required.
 * Be sure to call the appropriate "clear" function to free the memory.
 *-------------------------------------------------------------------------
 */
void set_db_username (DbLoginData *db_data, const char *username);
void set_db_password (DbLoginData *db_data, const char *password);
void set_db_name (DbLoginData *db_data, const char *name);
void set_db_descr (DbLoginData *db_data, const char *descr);
void set_db_host (DbLoginData *db_data, const char *host);
void set_db_port (DbLoginData *db_data, const char *port);

/*-------------------------------------------------------------------------
 * clear_db_XXXX
 *
 * Free the space used by the field.
 *-------------------------------------------------------------------------
 */
void clear_db_username (DbLoginData *db_data);
void clear_db_password (DbLoginData *db_data);
void clear_db_name (DbLoginData *db_data);
void clear_db_descr (DbLoginData *db_data);
void clear_db_host (DbLoginData *db_data);
void clear_db_port (DbLoginData *db_data);

/*-------------------------------------------------------------------------
 * Creation and Destruction.  If you create a DbLoginData structure,
 * be sure to free it.
 *-------------------------------------------------------------------------
 */
DbLoginData *db_login_data_new (void);
void db_login_data_free (DbLoginData *db_data);

/*-------------------------------------------------------------------------
 * Not sure what I want to do here...
 *-------------------------------------------------------------------------
 */
gboolean db_connect (DbLoginData *db_data);
void db_disconnect (DbLoginData *db_data);

/*-------------------------------------------------------------------------
 * Check for SQL errors, return FALSE if there is an error.
 *-------------------------------------------------------------------------
 */
gboolean chk_sql_ok (PGresult *res, const gchar *descr_text);

/*-------------------------------------------------------------------------
 * Drop the given database
 *-------------------------------------------------------------------------
 */
void drop_database (DbLoginData *pgsu_data, const gchar *db_name);
#endif
