/*-------------------------------------------------------------------------
 * Copyright (c) 2003-2005 Kenneth W. Sodemann (stuffle@mac.com)
 *-------------------------------------------------------------------------
 * db_schema
 *
 * Synopsis:
 *    Define the database schema.  If the schema is modified, the
 *    initialization functions in this file will need to be modified
 *    accordingly.
 *    
 * $Id: db_schema.c,v 1.20 2005/01/30 02:15:59 stuffle Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330 
 * Boston, MA  02111-1307  USA
 *-------------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "db_schema.h"

#define PERSON_RI "\
DECLARE\n\
   myperson RECORD;\n\
BEGIN\n\
   SELECT * INTO myperson\n\
     FROM person\n\
    WHERE person.login_id = NEW.login_id;\n\
   IF NOT FOUND THEN\n\
      RAISE EXCEPTION ''% does not exist.  Check person list.'',\n\
                      NEW.login_id;\n\
   END IF;\n\
   RETURN NEW;\n\
END;"

#define PROBLEM_REPORT_RI "\
DECLARE\n\
   myprob RECORD;\n\
BEGIN\n\
   SELECT * INTO myprob\n\
     FROM problem_report\n\
    WHERE problem_report.problem_num = NEW.problem_num;\n\
   IF NOT FOUND THEN\n\
      RAISE EXCEPTION ''Problem report % does not exist.'',\n\
          NEW.problem_num;\n\
   END IF;\n\
   RETURN NEW;\n\
END;"

#define PROBLEM_TYPE_RI "\
DECLARE\n\
   mytype RECORD;\n\
BEGIN\n\
   SELECT * INTO mytype\n\
     FROM problem_type\n\
    WHERE problem_type.type_num = NEW.type_num;\n\
   IF NOT FOUND THEN\n\
      RAISE EXCEPTION''Problem type % does not exist.'',\n\
          NEW.type_num;\n\
   END IF;\n\
   RETURN NEW;\n\
END;"

#define PROJECT_RI "\
DECLARE\n\
   myproj RECORD;\n\
BEGIN\n\
   SELECT * INTO myproj\n\
     FROM project\n\
    WHERE project.project_num = NEW.project_num;\n\
   IF NOT FOUND THEN\n\
      RAISE EXCEPTION ''Project number % does not exist.'',\n\
         NEW.project_num;\n\
   END IF;\n\
   RETURN NEW;\n\
END;"

#define SEVERITY_RI "\
DECLARE\n\
   mysever RECORD;\n\
BEGIN\n\
   SELECT * INTO mysever\n\
     FROM severity\n\
    WHERE severity.severity_num = NEW.severity_num;\n\
   IF NOT FOUND THEN\n\
      RAISE EXCEPTION ''Severity % does not exist.'', NEW.severity_num;\n\
   END IF;\n\
   RETURN NEW;\n\
END;"

#define STATUS_RI "\
DECLARE\n\
   mystatus RECORD;\n\
BEGIN\n\
   SELECT * INTO mystatus\n\
     FROM status\n\
    WHERE status.status_num = NEW.status_num;\n\
   IF NOT FOUND THEN\n\
      RAISE EXCEPTION ''Status % does not exist.'', NEW.status_num;\n\
   END IF;\n\
   RETURN NEW;\n\
END;"

#define AU_TR_STATUS_RI "\
DECLARE\n\
   mystatus RECORD;\n\
BEGIN\n\
   SELECT * INTO mystatus\n\
     FROM status\n\
    WHERE status.status_num = NEW.prev_state;\n\
   IF NOT FOUND THEN\n\
      RAISE DEBUG ''Previous status % no longer exists.'',\n\
         NEW.prev_state;\n\
   END IF;\n\
   SELECT * INTO mystatus\n\
     FROM status\n\
    WHERE status.status_num = NEW.new_state;\n\
   IF NOT FOUND THEN\n\
      RAISE DEBUG ''New status % does not exist.'', NEW.new_state;\n\
   END IF;\n\
   RETURN NEW;\n\
END;"

#define AU_TR_SEVR_RI "\
DECLARE\n\
   mysevr RECORD;\n\
BEGIN\n\
   SELECT * INTO mysevr\n\
     FROM severity\n\
    WHERE severity.severity_num = NEW.prev_sevr;\n\
   IF NOT FOUND THEN\n\
      RAISE DEBUG ''Previous severity % no longer exists.'',\n\
         NEW.prev_sevr;\n\
   END IF;\n\
   SELECT * INTO mysevr\n\
     FROM severity\n\
    WHERE severity.severity_num = NEW.new_sevr;\n\
   IF NOT FOUND THEN\n\
      RAISE DEBUG ''New severity % does not exist.'', NEW.new_sevr;\n\
   END IF;\n\
   RETURN NEW;\n\
END;"

#define CHECK_SUBMITTER "\
DECLARE\n\
   mysub RECORD;\n\
BEGIN\n\
   SELECT * INTO mysub\n\
     FROM submitter\n\
    WHERE submitter.project_num = NEW.project_num\n\
      AND submitter.login_id = NEW.submitter_id;\n\
   IF NOT FOUND THEN\n\
      RAISE EXCEPTION ''Submitter not valid for this project.'';\n\
   END IF;\n\
   RETURN NEW;\n\
END;"

#define CHECK_RESPONSIBLE_INS "\
DECLARE\n\
   myresp RECORD;\n\
BEGIN\n\
   IF NOT NEW.responsible_id ISNULL THEN\n\
      SELECT * INTO myresp\n\
        FROM responsible\n\
       WHERE responsible.project_num = NEW.project_num\n\
         AND responsible.login_id = NEW.responsible_id;\n\
      IF NOT FOUND THEN\n\
         RAISE EXCEPTION ''Responsible person not valid for this project.'';\n\
      END IF;\n\
   END IF;\n\
   RETURN NEW;\n\
END;"

#define CHECK_RESPONSIBLE_UPD "\
DECLARE\n\
   myresp RECORD;\n\
BEGIN\n\
   IF NOT NEW.responsible_id ISNULL AND\n\
      (OLD.responsible_id ISNULL OR\n\
       OLD.responsible_id <> NEW. responsible_id)\n\
   THEN\n\
      SELECT * INTO myresp\n\
        FROM responsible\n\
       WHERE responsible.project_num = NEW.project_num\n\
         AND responsible.login_id = NEW.responsible_id;\n\
      IF NOT FOUND THEN\n\
         RAISE EXCEPTION ''Responsible person not valid for this project.'';\n\
      END IF;\n\
   END IF;\n\
   RETURN NEW;\n\
END;"

#define CHECK_ACCESS "\
DECLARE\n\
   lclrec  RECORD;\n\
BEGIN\n\
   SELECT *\n\
     INTO lclrec\n\
     FROM person\n\
    WHERE person.login_id = getpgusername()\n\
      AND person.access_flag = 1;\n\
   IF NOT FOUND THEN\n\
      RAISE EXCEPTION ''Access Denied!!'';\n\
   END IF;\n\
   RETURN NEW;\n\
END;"

#define CHECK_ACCESS_DEL "\
DECLARE\n\
   lclrec  RECORD;\n\
BEGIN\n\
   SELECT *\n\
     INTO lclrec\n\
     FROM person\n\
    WHERE person.login_id = getpgusername()\n\
      AND person.access_flag = 1;\n\
   IF NOT FOUND THEN\n\
      RAISE EXCEPTION ''Access Denied!!'';\n\
   END IF;\n\
   RETURN OLD;\n\
END;"

#define CHECK_QUERY_ACCESS "\
DECLARE\n\
   lclrec RECORD;\n\
BEGIN\n\
   IF NEW.is_system_query = TRUE THEN\n\
      SELECT *\n\
        INTO lclrec\n\
        FROM person\n\
       WHERE login_id = getpgusername ()\n\
         AND access_flag = 1;\n\
      IF NOT FOUND THEN\n\
         RAISE EXCEPTION ''Must be admin to edit system queries'';\n\
      END IF;\n\
   END IF;\n\
   RETURN NEW;\n\
END;"

#define CHECK_QUERY_ACCESS_DEL "\
DECLARE\n\
   lclrec  RECORD;\n\
BEGIN\n\
   IF OLD.is_system_query = TRUE THEN\n\
      SELECT *\n\
        INTO lclrec\n\
        FROM person\n\
       WHERE login_id = getpgusername ()\n\
         AND access_flag = 1;\n\
      IF NOT FOUND THEN\n\
         RAISE EXCEPTION ''Must be admin to delete system queries'';\n\
      END IF;\n\
   END IF;\n\
   RETURN OLD;\n\
END;"

#define CHECK_LAST_ADMIN "\
DECLARE\n\
   cnt int4;\n\
BEGIN\n\
   -- We really are only concerned about cases where\n\
   -- we are changing a user from admin to non-admin\n\
   IF OLD.access_flag = 1 AND NEW.access_flag <> 1 THEN\n\
      SELECT count (*)\n\
        INTO cnt\n\
        FROM person\n\
       WHERE person.access_flag = 1;\n\
      IF cnt = 1 THEN\n\
         RAISE EXCEPTION ''Changing access type would leave no admins.'';\n\
      END IF;\n\
   END IF;\n\
   RETURN NEW;\n\
END;"

#define PERSON_DEL_CLEANUP "\
DECLARE\n\
   lclrec  RECORD;\n\
   cnt     int4;\n\
BEGIN\n\
   -- Does the current user have access to delete?\n\
   SELECT *\n\
     INTO lclrec\n\
     FROM person\n\
    WHERE person.login_id = getpgusername()\n\
      AND person.access_flag = 1;\n\
   IF NOT FOUND THEN\n\
      RAISE EXCEPTION ''Access Denied!!'';\n\
   END IF;\n\
   -- Lets see if the user is used at all\n\
   SELECT *\n\
     INTO lclrec\n\
     FROM project\n\
    WHERE project.login_id = OLD.login_id;\n\
   IF FOUND THEN\n\
      RAISE EXCEPTION ''User leads at least one project.  Cannot delete.'';\n\
   END IF;\n\
   SELECT *\n\
     INTO lclrec\n\
     FROM problem_report\n\
    WHERE responsible_id = OLD.login_id\n\
       OR submitter_id = OLD.login_id;\n\
   IF FOUND THEN\n\
      RAISE EXCEPTION ''User has submitted or is reponsible for PRs.'';\n\
   END IF;\n\
   -- If this is the last admin user, do not allow change.\n\
   IF OLD.access_flag = 1 THEN\n\
      SELECT count (*)\n\
        INTO cnt\n\
        FROM person\n\
       WHERE person.access_flag = 1;\n\
      IF cnt = 1 THEN\n\
         RAISE EXCEPTION ''Cannot delete last user with admin access.'';\n\
      END IF;\n\
   END IF;\n\
   -- If we made it this far, we are going all the way.  Delete any\n\
   -- rows in the submitter or reponsible table for this user.\n\
   DELETE FROM responsible\n\
    WHERE responsible.login_id = OLD.login_id;\n\
   DELETE FROM submitter\n\
    WHERE submitter.login_id = OLD.login_id;\n\
   RETURN OLD;\n\
END;"

#define PR_INSERT_ACTIVE "\
DECLARE\n\
   active_flag boolean;\n\
   name_text   text;\n\
BEGIN\n\
   SELECT active, name INTO active_flag, name_text\n\
     FROM severity\n\
    WHERE severity.severity_num = NEW.severity_num;\n\
   IF FOUND AND NOT active_flag THEN\n\
      RAISE EXCEPTION ''Severity \"%\" is not active.'', name_text;\n\
   END IF;\n\
   SELECT active, name INTO active_flag, name_text\n\
     FROM status\n\
    WHERE status.status_num = NEW.status_num;\n\
   IF FOUND AND NOT active_flag THEN\n\
      RAISE EXCEPTION ''Status \"%\" is not active.'', name_text;\n\
   END IF;\n\
   SELECT active, name INTO active_flag, name_text\n\
     FROM problem_type\n\
    WHERE problem_type.type_num = NEW.type_num;\n\
   IF FOUND AND NOT active_flag THEN\n\
      RAISE EXCEPTION ''Type \"%\" is not active.'', name_text;\n\
   END IF;\n\
   SELECT project.active, project.name\n\
     INTO active_flag, name_text\n\
     FROM project\n\
    WHERE project.project_num = NEW.project_num;\n\
   IF FOUND AND NOT active_flag THEN\n\
      RAISE EXCEPTION ''Project \"%\" is not active.'', name_text;\n\
   END IF;\n\
   RETURN NEW;\n\
END;"

#define PR_UPDATE_ACTIVE "\
DECLARE\n\
   active_flag boolean;\n\
   name_text   text;\n\
BEGIN\n\
   IF NEW.status_num != OLD.status_num THEN\n\
      SELECT active, name INTO active_flag, name_text\n\
        FROM status\n\
       WHERE status.status_num = NEW.status_num;\n\
      IF FOUND AND NOT active_flag THEN\n\
         RAISE EXCEPTION ''Status \"%\" is not active'',\n\
             name_text;\n\
      END IF;\n\
   END IF;\n\
   IF NEW.severity_num != OLD.severity_num THEN\n\
      SELECT active, name INTO active_flag, name_text\n\
        FROM severity\n\
       WHERE severity.severity_num = NEW.severity_num;\n\
      IF FOUND AND NOT active_flag THEN\n\
         RAISE EXCEPTION ''Severity \"%\" is not active'',\n\
            name_text;\n\
      END IF;\n\
   END IF;\n\
   IF NEW.type_num != OLD.type_num THEN\n\
      RAISE EXCEPTION ''You cannot change the type on a problem.'';\n\
   END IF;\n\
   IF NEW.project_num != OLD.project_num THEN\n\
      RAISE EXCEPTION ''You cannot change the project on a problem.'';\n\
   END IF;\n\
   RETURN NEW;\n\
END;"

#define CREATION_STAMP "\
BEGIN\n\
   NEW.creator = getpgusername();\n\
   NEW.creation_date = now();\n\
   RETURN NEW;\n\
END;"

#define SUBMITTER_NO_CHANGE "\
BEGIN\n\
   IF OLD.submitter_id != NEW.submitter_id THEN\n\
      RAISE EXCEPTION ''Cannot change the submitter!!'';\n\
   END IF;\n\
   RETURN NEW;\n\
END;"

#define CHECK_SEVERITY_NOT_USED "\
DECLARE\n\
   cnt int4;\n\
BEGIN\n\
   SELECT COUNT (*) INTO cnt\n\
     FROM problem_type pt\n\
    WHERE pt.severity_num = OLD.severity_num;\n\
   IF cnt <> 0 THEN\n\
      RAISE EXCEPTION\n\
         ''Severity: %, is used as a default for one or more PR types'',\n\
         OLD.name;\n\
   END IF;\n\
   SELECT COUNT (*) INTO cnt\n\
     FROM problem_report pr\n\
    WHERE pr.severity_num = OLD.severity_num;\n\
   IF cnt <> 0 THEN\n\
      RAISE EXCEPTION ''Severity: %, is used in on or more problem reports'',\n\
         OLD.name;\n\
   END IF;\n\
   RETURN OLD;\n\
END;"

#define CHECK_STATUS_NOT_USED "\
DECLARE\n\
   cnt int4;\n\
BEGIN\n\
   SELECT COUNT (*) INTO cnt\n\
     FROM problem_report pr\n\
    WHERE pr.status_num = OLD.status_num;\n\
   IF cnt <> 0 THEN\n\
      RAISE EXCEPTION\n\
         ''Status: %, is used in one or more problem reports'',\n\
         OLD.name;\n\
   END IF;\n\
   RETURN OLD;\n\
END;"

#define CHECK_TYPE_NOT_USED "\
DECLARE\n\
   cnt int4;\n\
BEGIN\n\
   SELECT COUNT (*) INTO cnt\n\
     FROM problem_report pr\n\
    WHERE pr.type_num = OLD.type_num;\n\
   IF cnt <> 0 THEN\n\
      RAISE EXCEPTION\n\
         ''Type: %, is used in one or more problem reports'',\n\
         OLD.name;\n\
   END IF;\n\
   RETURN OLD;\n\
END;"

#define NEW_PR_EMAIL_LIST "\
DECLARE\n\
   email_list  TEXT;\n\
   tmp_rec     RECORD;\n\
   prj_num     ALIAS FOR $1;\n\
BEGIN\n\
   email_list := NULL;\n\
   FOR tmp_rec IN\n\
      SELECT DISTINCT p.e_mail\n\
        FROM person p, project prj\n\
       WHERE ((p.login_id IN (SELECT r.login_id\n\
                                FROM responsible r\n\
                               WHERE r.project_num = prj.project_num) AND\n\
               prj.new_pr_notify_grp = TRUE) OR\n\
              (p.login_id = prj.login_id AND\n\
               prj.new_pr_notify_pl = TRUE))\n\
         AND prj.project_num = prj_num\n\
   LOOP\n\
      IF NOT tmp_rec.e_mail ISNULL THEN\n\
         IF email_list ISNULL THEN\n\
            email_list := tmp_rec.e_mail;\n\
         ELSE\n\
            email_list := textcat( email_list, '' '' );\n\
            email_list := textcat( email_list, tmp_rec.e_mail );\n\
         END IF;\n\
      END IF;\n\
   END LOOP;\n\
   RETURN email_list;\n\
END;"

#define NEW_PR_SEND_NOTIFICATION "\
DECLARE\n\
   email_list        TEXT;\n\
   project_name      TEXT;\n\
   pr_title          TEXT;\n\
   descr             TEXT;\n\
   severity_name     TEXT;\n\
   problem_type_name TEXT;\n\
   submitter_name    TEXT;\n\
   bogus             int4;\n\
BEGIN\n\
   SELECT new_pr_email_list( NEW.project_num )\n\
     INTO email_list;\n\
   SELECT ''PR #'' || NEW.problem_num || '': '' || NEW.title,\n\
          NEW.descr, sevr.name, pt.name, prj.name,\n\
          p1.last_name || '', '' || p1.first_name || '' <'' ||\n\
          p1.e_mail || ''>''\n\
     INTO pr_title, descr, severity_name, problem_type_name,\n\
          project_name, submitter_name\n\
     FROM severity sevr, person p1, project prj, problem_type pt\n\
    WHERE prj.project_num = NEW.project_num\n\
      AND sevr.severity_num = NEW.severity_num\n\
      AND pt.type_num = NEW.type_num\n\
      AND p1.login_id = NEW.submitter_id;\n\
      IF NOT email_list ISNULL THEN\n\
      SELECT send_new_pr_msg( email_list, project_name, pr_title,\n\
                              descr, severity_name, problem_type_name,\n\
                              submitter_name )\n\
        INTO bogus;\n\
   END IF;\n\
   return NEW;\n\
END;"

#define ASSIGN_PR_EMAIL_LIST "\
DECLARE\n\
   email_list   TEXT;\n\
   tmp_rec      RECORD;\n\
   prj_num      ALIAS FOR $1;\n\
   resp_id      ALIAS FOR $2;\n\
   sub_id       ALIAS FOR $3;\n\
BEGIN\n\
   email_list := NULL;\n\
   FOR tmp_rec IN\n\
   SELECT DISTINCT p.e_mail\n\
     FROM person p, project prj\n\
    WHERE ((p.login_id = prj.login_id AND\n\
            prj.assign_notify_pl = TRUE) OR\n\
           (p.login_id = resp_id AND\n\
            prj.assign_notify_resp = TRUE) OR\n\
           (p.login_id = sub_id AND\n\
            prj.assign_notify_sub = TRUE))\n\
      AND prj.project_num = prj_num\n\
   LOOP\n\
      IF NOT tmp_rec.e_mail ISNULL THEN\n\
         IF email_list ISNULL THEN\n\
            email_list := tmp_rec.e_mail;\n\
         ELSE\n\
            email_list := textcat( email_list, '' '' );\n\
            email_list := textcat( email_list, tmp_rec.e_mail );\n\
         END IF;\n\
      END IF;\n\
   END LOOP;\n\
   RETURN email_list;\n\
END;"

#define UPDATE_PR_EMAIL_LIST "\
DECLARE\n\
   email_list   TEXT;\n\
   tmp_rec      RECORD;\n\
   prj_num      ALIAS FOR $1;\n\
   resp_id      ALIAS FOR $2;\n\
   sub_id       ALIAS FOR $3;\n\
BEGIN\n\
   email_list = NULL;\n\
   FOR tmp_rec IN\n\
   SELECT DISTINCT p.e_mail\n\
     FROM person p, project prj\n\
    WHERE ((p.login_id IN (SELECT r.login_id\n\
                             FROM responsible r\n\
                            WHERE r.project_num = prj.project_num) AND\n\
            prj.upd_notify_grp = TRUE) OR\n\
           (p.login_id = prj.login_id AND\n\
            prj.upd_notify_pl = TRUE) OR\n\
           (p.login_id = resp_id AND\n\
            prj.upd_notify_resp = TRUE) OR\n\
           (p.login_id = sub_id AND\n\
            prj.upd_notify_sub = TRUE))\n\
      AND prj.project_num = prj_num\n\
   LOOP\n\
      IF NOT tmp_rec.e_mail ISNULL THEN\n\
         IF email_list ISNULL THEN\n\
            email_list := tmp_rec.e_mail;\n\
         ELSE\n\
            email_list := textcat( email_list, '' '' );\n\
            email_list := textcat( email_list, tmp_rec.e_mail );\n\
         END IF;\n\
      END IF;\n\
   END LOOP;\n\
   RETURN email_list;\n\
END;"

#define PRJ_VERSION_RI "\
DECLARE\n\
   my_version  RECORD;\n\
   check_open  BOOLEAN;\n\
   check_close BOOLEAN;\n\
BEGIN\n\
   check_open := NOT NEW.open_version ISNULL;\n\
   check_close := NOT NEW.close_version ISNULL;\n\
   IF TG_OP = ''UPDATE'' THEN\n\
      check_open := check_open AND (NEW.open_version <> OLD.open_version);\n\
      check_close :=\n\
         check_close AND (NEW.close_version <> OLD.close_version);\n\
   END IF;\n\
   IF check_open THEN\n\
      SELECT *\n\
        INTO my_version\n\
        FROM project_version\n\
       WHERE version_num = NEW.open_version\n\
         AND active = TRUE;\n\
      IF NOT FOUND THEN\n\
         RAISE EXCEPTION ''Open version % does not exist or is inactive.'',\n\
            NEW.open_version;\n\
      END IF;\n\
   END IF;\n\
   IF check_close THEN\n\
      SELECT *\n\
        INTO my_version\n\
        FROM project_version\n\
       WHERE version_num = NEW.close_version\n\
         AND active = TRUE;\n\
      IF NOT FOUND THEN\n\
         RAISE EXCEPTION ''Close version % does not exist or is inactive.'',\n\
            NEW.close_version;\n\
      END IF;\n\
   END IF;\n\
   RETURN NEW;\n\
END;"

#define CHECK_CLOSE_VERSION "\
DECLARE\n\
   my_record  RECORD;\n\
   ver        project.close_version%TYPE;\n\
BEGIN\n\
   IF NOT NEW.open_version ISNULL AND\n\
          NEW.close_version ISNULL THEN\n\
      SELECT *\n\
        INTO my_record\n\
        FROM status\n\
       WHERE status_num = NEW.status_num\n\
         AND is_closed_status = TRUE;\n\
      IF FOUND THEN\n\
         SELECT close_version\n\
           INTO ver\n\
           FROM project\n\
          WHERE project_num = NEW.project_num;\n\
         IF ver ISNULL THEN\n\
            RAISE EXCEPTION ''Must supply a closing version.'';\n\
         END IF;\n\
         NEW.close_version := ver;\n\
      END IF;\n\
   END IF;\n\
   RETURN NEW;\n\
END;"

#define CHECK_PRJ_VERSION_NOT_USED "\
DECLARE\n\
   cnt int4;\n\
BEGIN\n\
   SELECT COUNT (*)\n\
     INTO cnt\n\
     FROM problem_report\n\
    WHERE (open_version = OLD.version_num OR\n\
           close_version = OLD.version_num)\n\
      AND project_num = OLD.project_num;\n\
   IF cnt <> 0 THEN\n\
      RAISE EXCEPTION ''Version % is used in one or more problem reports.'',\n\
         OLD.version_text;\n\
   END IF;\n\
   SELECT COUNT (*)\n\
     INTO cnt\n\
     FROM project\n\
    WHERE (open_version = OLD.version_num OR\n\
           close_version = OLD.version_num)\n\
      AND project_num = OLD.project_num;\n\
   IF cnt <> 0 THEN\n\
      RAISE EXCEPTION ''Version % is used as a default for this project.'',\n\
         OLD.version_text;\n\
   END IF;\n\
   RETURN OLD;\n\
END;"

#define NEXT_VERSION_NUM "\
DECLARE\n\
   prj_num ALIAS for $1;\n\
   ver_num int4;\n\
BEGIN\n\
   SELECT max( version_num ) + 1\n\
     INTO ver_num\n\
     FROM project_version\n\
    WHERE project_num = prj_num;\n\
   IF ver_num ISNULL THEN\n\
      ver_num := 1;\n\
   END IF;\n\
   RETURN ver_num;\n\
END;"

#define CLEAN_UP_PROJECT "\
DECLARE\n\
   lclrec  RECORD;\n\
BEGIN\n\
   -- First see if the user has permission to do this\n\
   -- operation.\n\
   SELECT *\n\
     INTO lclrec\n\
     FROM person, project\n\
    WHERE project.login_id = getpgusername()\n\
       OR (person.login_id = getpgusername()\n\
           AND person.access_flag = 1);\n\
   IF NOT FOUND THEN\n\
      RAISE EXCEPTION ''Access Denied!!'';\n\
   END IF;\n\
   -- Then delete all of the associated problem reports.\n\
   DELETE\n\
     FROM problem_report\n\
    WHERE project_num = OLD.project_num;\n\
   -- Clean up the responsible and submitter tables also.\n\
   DELETE\n\
     FROM submitter\n\
    WHERE project_num = OLD.project_num;\n\
   DELETE\n\
     FROM responsible\n\
     WHERE project_num = OLD.project_num;\n\
   RETURN OLD;\n\
END;"

#define CLEAN_UP_PROJECT_V3 "\
DECLARE\n\
   lclrec  RECORD;\n\
BEGIN\n\
   -- First see if the user has permission to do this\n\
   -- operation.\n\
   SELECT *\n\
     INTO lclrec\n\
     FROM person, project\n\
    WHERE project.login_id = getpgusername()\n\
       OR (person.login_id = getpgusername()\n\
           AND person.access_flag = 1);\n\
   IF NOT FOUND THEN\n\
      RAISE EXCEPTION ''Access Denied!!'';\n\
   END IF;\n\
   -- Then delete all of the associated problem reports.\n\
   DELETE\n\
     FROM problem_report\n\
    WHERE project_num = OLD.project_num;\n\
   -- Clean up the responsible and submitter tables also.\n\
   DELETE\n\
     FROM submitter\n\
    WHERE project_num = OLD.project_num;\n\
   DELETE\n\
     FROM responsible\n\
     WHERE project_num = OLD.project_num;\n\
   -- Clean up the project version table\n\
   DELETE\n\
     FROM project_version\n\
    WHERE project_num = OLD.project_num;\n\
   RETURN OLD;\n\
END;"

#define AUTO_ASSIGN_PR "\
DECLARE\n\
   user_id   project.auto_assign_login_id%TYPE;\n\
BEGIN\n\
   SELECT auto_assign_login_id\n\
     INTO user_id\n\
     FROM project\n\
    WHERE project_num = NEW.project_num;\n\
   IF NOT user_id ISNULL THEN\n\
      NEW.responsible_id := user_id;\n\
   END IF;\n\
   RETURN NEW;\n\
END;"

#define AUTO_ASSIGN_RI "\
DECLARE\n\
   resp  RECORD;\n\
BEGIN\n\
   SELECT *\n\
     INTO resp\n\
     FROM responsible\n\
    WHERE project_num = NEW.project_num\n\
      AND login_id = NEW.auto_assign_login_id;\n\
   IF NOT FOUND AND NOT NEW.auto_assign_login_id ISNULL THEN\n\
      RAISE EXCEPTION ''% is not a project member.'',\n\
         NEW.auto_assign_login_id;\n\
   END IF;\n\
   RETURN NEW;\n\
END;"

#define CHECK_QUERY_NAME "\
DECLARE\n\
   rec  RECORD;\n\
BEGIN\n\
   SELECT *\n\
     INTO rec\n\
     FROM pr_query\n\
    WHERE name = NEW.name\n\
      AND query_num <> NEW.query_num\n\
      AND is_system_query = TRUE;\n\
   IF FOUND THEN\n\
      RAISE EXCEPTION ''A system query named \"%\" already exists.'',\n\
         NEW.name;\n\
   END IF;\n\
   SELECT *\n\
     INTO rec\n\
     FROM pr_query\n\
    WHERE name = NEW.name\n\
      AND query_num <> NEW.query_num\n\
      AND is_system_query = FALSE\n\
      AND NEW.is_system_query = TRUE;\n\
   IF FOUND THEN\n\
      RAISE EXCEPTION ''User % already has a query named \"%\".'',\n\
         rec.creator, NEW.name;\n\
   END IF;\n\
   SELECT *\n\
     INTO rec\n\
     FROM pr_query\n\
    WHERE name = NEW.name\n\
      AND query_num <> NEW.query_num\n\
      AND is_system_query = FALSE\n\
      AND NEW.is_system_query = FALSE\n\
      AND creator = getpgusername();\n\
   IF FOUND THEN\n\
      RAISE EXCEPTION ''You already have a query named \"%\".'', NEW.name;\n\
   END IF;\n\
   RETURN NEW;\n\
END;"

#define CAN_EDIT_PROJECT "\
DECLARE\n\
   lclrec    RECORD;\n\
   rtn_val   boolean;\n\
   prj_num   ALIAS FOR $1;\n\
BEGIN\n\
   SELECT *\n\
     INTO lclrec\n\
     FROM person, project\n\
    WHERE (project.login_id = getpgusername()\n\
           OR (person.login_id = getpgusername()\n\
               AND person.access_flag = 1))\n\
      AND project.project_num = prj_num;\n\
   IF NOT FOUND THEN\n\
      rtn_val := FALSE;\n\
   ELSE\n\
      rtn_val := TRUE;\n\
   END IF;\n\
   RETURN rtn_val;\n\
END;"

#define CHECK_PRJ_ACCESS "\
DECLARE\n\
    lclrec  RECORD;\n\
BEGIN\n\
    SELECT *\n\
      INTO lclrec\n\
      FROM person, project\n\
     WHERE project.login_id = getpgusername()\n\
        OR (person.login_id = getpgusername()\n\
            AND person.access_flag = 1);\n\
    IF NOT FOUND THEN\n\
        RAISE EXCEPTION ''Access Denied!!'';\n\
    END IF;\n\
    RETURN NEW;\n\
END;"

#define CHECK_PRJ_ACCESS_V4 "\
DECLARE\n\
   can_edit  boolean;\n\
BEGIN\n\
   SELECT can_edit_project (NEW.project_num)\n\
     INTO can_edit;\n\
   IF NOT can_edit THEN\n\
      RAISE EXCEPTION ''Access Denied!!'';\n\
   END IF;\n\
   RETURN NEW;\n\
END;"

#define ASSIGN_SEND_NOTIFICATION "\
DECLARE\n\
   email_list        TEXT;\n\
   pr_title          TEXT;\n\
   descr             TEXT;\n\
   fix_descr         TEXT;\n\
   status_name       TEXT;\n\
   severity_name     TEXT;\n\
   responsible_name  TEXT;\n\
   submitter_name    TEXT;\n\
   bogus             int4;\n\
BEGIN\n\
   if nameeq(NEW.responsible_id, OLD.responsible_id) then\n\
      return NEW;\n\
   end if;\n\
   if NEW.responsible_id ISNULL and\n\
      OLD.responsible_id ISNULL then\n\
      return NEW;\n\
   end if;\n\
   SELECT assign_pr_email_list (NEW.project_num,\n\
                                NEW.responsible_id,\n\
                                NEW.submitter_id)\n\
     INTO email_list;\n\
   IF NOT email_list ISNULL THEN\n\
      if NEW.responsible_id ISNULL then\n\
         SELECT ''PR #'' || NEW.problem_num || '': '' || NEW.title,\n\
                NEW.descr, NEW.fix, stat.name, sevr.name,\n\
                ''no one'',\n\
                p2.last_name || '', '' || p2.first_name || '' <'' ||\n\
                p2.e_mail || ''>''\n\
           INTO pr_title, descr, fix_descr, status_name,\n\
               severity_name, responsible_name, submitter_name\n\
          FROM status stat, severity sevr,  person p2\n\
         WHERE stat.status_num = NEW.status_num\n\
           AND sevr.severity_num = NEW.severity_num\n\
           AND p2.login_id = NEW.submitter_id;\n\
      else\n\
         SELECT ''PR #'' || NEW.problem_num || '': '' || NEW.title,\n\
                NEW.descr, NEW.fix, stat.name, sevr.name,\n\
                p1.last_name || '', '' || p1.first_name || '' <'' ||\n\
               p1.e_mail || ''>'',\n\
               p2.last_name || '', '' || p2.first_name || '' <'' ||\n\
               p2.e_mail || ''>''\n\
          INTO pr_title, descr, fix_descr, status_name,\n\
               severity_name, responsible_name, submitter_name\n\
          FROM status stat, severity sevr, person p1, person p2\n\
         WHERE stat.status_num = NEW.status_num\n\
           AND sevr.severity_num = NEW.severity_num\n\
           AND p1.login_id = NEW.responsible_id\n\
           AND p2.login_id = NEW.submitter_id;\n\
      end if;\n\
      SELECT send_assign_msg( email_list, pr_title, descr, fix_descr,\n\
                              status_name, severity_name,\n\
                              responsible_name, submitter_name )\n\
        INTO bogus;\n\
   END IF;\n\
   return NEW;\n\
END;"

#define UPDATE_SEND_NOTIFICATION "\
DECLARE\n\
   email_list        TEXT;\n\
   pr_title          TEXT;\n\
   descr             TEXT;\n\
   fix_descr         TEXT;\n\
   status_name       TEXT;\n\
   severity_name     TEXT;\n\
   responsible_name  TEXT;\n\
   submitter_name    TEXT;\n\
   bogus             int4;\n\
BEGIN\n\
   -- If the status has not changed, just return\n\
   if NEW.status_num = OLD.status_num then\n\
      return NEW;\n\
   end if;\n\
   SELECT update_pr_email_list (NEW.project_num,\n\
                                NEW.responsible_id,\n\
                                NEW.submitter_id)\n\
     INTO email_list;\n\
   IF NOT email_list ISNULL THEN\n\
      if NEW.responsible_id ISNULL then\n\
         SELECT ''PR #'' || NEW.problem_num || '': '' || NEW.title,\n\
                NEW.descr, NEW.fix, stat.name, sevr.name,\n\
                ''No One'',\n\
                 p2.last_name || '', '' || p2.first_name || '' <'' ||\n\
                p2.e_mail || ''>''\n\
           INTO pr_title, descr, fix_descr, status_name, severity_name,\n\
               responsible_name, submitter_name\n\
          FROM status stat, severity sevr,  person p2\n\
         WHERE stat.status_num = NEW.status_num\n\
           AND sevr.severity_num = NEW.severity_num\n\
           AND p2.login_id = NEW.submitter_id;\n\
      else\n\
         SELECT ''PR #'' || NEW.problem_num || '': '' || NEW.title,\n\
                NEW.descr, NEW.fix, stat.name, sevr.name,\n\
                p1.last_name || '', '' || p1.first_name || '' <'' ||\n\
                p1.e_mail || ''>'',\n\
                p2.last_name || '', '' || p2.first_name || '' <'' ||\n\
                p2.e_mail || ''>''\n\
        INTO pr_title, descr, fix_descr, status_name, severity_name,\n\
             responsible_name, submitter_name\n\
        FROM status stat, severity sevr, person p1, person p2\n\
       WHERE stat.status_num = NEW.status_num\n\
         AND sevr.severity_num = NEW.severity_num\n\
         AND p1.login_id = NEW.responsible_id\n\
         AND p2.login_id = NEW.submitter_id;\n\
      end if;\n\
      SELECT send_update_msg( email_list, pr_title, descr, fix_descr,\n\
                              status_name, severity_name,\n\
                              responsible_name, submitter_name )\n\
        INTO bogus;\n\
   END IF;\n\
   return NEW;\n\
END;"


#define LOG_RULE_WHERE "((new.severity_num <> old.severity_num) OR \
(new.status_num <> old.status_num))"

#define LOG_RULE_ACTION "\
INSERT INTO audit_trail (the_date, problem_num, prev_state, new_state, \
prev_sevr, new_sevr, login_id) VALUES (now(), new.problem_num, \
old.status_num, new.status_num, old.severity_num, new.severity_num, \
getpgusername())"


#define INIT_ORDER_NUM "\
UPDATE %s \
   SET order_num = (select count (*) + 1 \
                      from %s as t\
                     where t.%s < %s.%s) \
 WHERE order_num is NULL"


static GList *sequences = NULL;
static GList *tables = NULL;
static GList *languages = NULL;
static GList *functions = NULL;
static GList *stored_funcs = NULL;
static GList *unique_indexes = NULL;
static GList *triggers = NULL;
static GList *rules = NULL;


static gint
count_list_elements (GList *list)
{
   GList *iter;
   gint   n = 0;

   for (iter = g_list_first (list);
         iter != NULL;
         iter = iter ->next)
      {
      n++;
      }

   return n;
}


gchar *
pg_type_name (PgDataType type)
{
   switch (type)
      {
      case PG_BOOL:
         return "bool";
      case PG_INT4:
         return "int4";
      case PG_DATE:
         return "date";
      case PG_TIMESTAMP:
         return "timestamp";
      case PG_CHRVAR:
         return "character varying";
      case PG_NAME:
         return "name";
      case PG_TEXT:
         return "text";
      case PG_INT4_ARR:
         return "\"_int4\"";
      case PG_NAME_ARR:
         return "\"_name\"";
      default:
         g_assert_not_reached ();
      }

   return "undefined";
}


/*
 * Sequence related functions.
 */
static void
initialize_sequences (void)
{
   SequenceData *seq;

   g_assert (sequences == NULL);

   seq = g_malloc (sizeof (SequenceData));
   seq->name = g_string_new ("status_seq");
   seq->ref_table_name = g_string_new ("status");
   seq->ref_col_name = g_string_new ("status_num");
   seq->start = 1;
   seq->increment = 1;
   seq->minval = 1;
   seq->maxval = 2147483647;
   seq->cache = 1;
   sequences = g_list_append (sequences, seq);

   seq = g_malloc (sizeof (SequenceData));
   seq->name = g_string_new ("severity_seq");
   seq->ref_table_name = g_string_new ("severity");
   seq->ref_col_name = g_string_new ("severity_num");
   seq->start = 1;
   seq->increment = 1;
   seq->minval = 1;
   seq->maxval = 2147483647;
   seq->cache = 1;
   sequences = g_list_append (sequences, seq);

   seq = g_malloc (sizeof (SequenceData));
   seq->name = g_string_new ("problem_type_seq");
   seq->ref_table_name = g_string_new ("problem_type");
   seq->ref_col_name = g_string_new ("problem_type_num");
   seq->start = 1;
   seq->increment = 1;
   seq->minval = 1;
   seq->maxval = 2147483647;
   seq->cache = 1;
   sequences = g_list_append (sequences, seq);

   seq = g_malloc (sizeof (SequenceData));
   seq->name = g_string_new ("project_seq");
   seq->ref_table_name = g_string_new ("project");
   seq->ref_col_name = g_string_new ("project_num");
   seq->start = 1;
   seq->increment = 1;
   seq->minval = 1;
   seq->maxval = 2147483647;
   seq->cache = 1;
   sequences = g_list_append (sequences, seq);

   seq = g_malloc (sizeof (SequenceData));
   seq->name = g_string_new ("problem_report_seq");
   seq->ref_table_name = g_string_new ("problem_report");
   seq->ref_col_name = g_string_new ("problem_num");
   seq->start = 1;
   seq->increment = 1;
   seq->minval = 1;
   seq->maxval = 2147483647;
   seq->cache = 1;
   sequences = g_list_append (sequences, seq);

   seq = g_malloc (sizeof (SequenceData));
   seq->name = g_string_new ("pr_query_seq");
   seq->ref_table_name = g_string_new ("pr_query");
   seq->ref_col_name = g_string_new ("query_num");
   seq->start = 1;
   seq->increment = 1;
   seq->minval = 1;
   seq->maxval = 2147483647;
   seq->cache = 1;
   sequences = g_list_append (sequences, seq);
}

static void
free_sequence (gpointer data, gpointer user_data)
{
   SequenceData *seq = data;

   g_string_free (seq->name, TRUE);
   g_string_free (seq->ref_table_name, TRUE);
   g_string_free (seq->ref_col_name, TRUE);
   g_free (seq);
}

static void
destroy_sequences (void)
{
   g_list_foreach (sequences, free_sequence, NULL);
   g_list_free (sequences);
   sequences = NULL;
}

gint
count_sequences (void)
{
   return count_list_elements (sequences);
}

SequenceData *
get_sequence (gint n)
{
   return g_list_nth_data (sequences, n);
}


/*
 * Table related functions
 */
static TableData *
new_table_data (const gchar *name, gboolean prereq, gint minver)
{
   TableData *tab;

   tab = g_malloc (sizeof (TableData));
   tab->name = g_string_new (name);
   tab->columns = NULL;
   tab->pk_cols = NULL;
   tab->constraints = NULL;
   tab->prereq = prereq;
   tab->minver = minver;

   return tab;
}

static void
add_cons_to_table (TableData   *tab,
                   const gchar *name,
                   const gchar *cond)
{
   ConstraintData *cons;

   cons = g_malloc (sizeof (ConstraintData));
   cons->name = g_string_new (name);
   cons->condition = g_string_new (cond);
   tab->constraints = g_list_append (tab->constraints, cons);
}

static ColumnData *
add_column_to_table (TableData   *tab,
                     const gchar *name,
                     PgDataType   type,
                     guint        size,
                     gboolean     nullable,
                     gboolean     key_col,
                     gint         minver)
{
   ColumnData *col;

   col = g_malloc (sizeof (ColumnData));
   col->name = g_string_new (name);
   col->type = type;
   col->size = size;
   col->nullable = nullable;
   /*
    * Assume that if a table exists, it must be complete.  If there is 
    * a column that can be added after the fact, it will likely need
    * to have initialization SQL.  Both of these items can be overridden
    * by the caller.
    */
   col->prereq = TRUE;
   col->initsql = NULL;
   col->minver = minver;
   tab->columns = g_list_append (tab->columns, col);
   if (key_col)
      {
      tab->pk_cols = g_list_append (tab->pk_cols, col);
      }

   return col;
}

static void
initialize_tables (void)
{
   TableData  *tab;
   ColumnData *col;

   g_assert (tables == NULL);

   /*
    * version
    */
   tab = new_table_data ("version", TRUE, 1);
   add_column_to_table (tab, "schema_ver",   PG_INT4, 0, FALSE, FALSE, 1);
   add_column_to_table (tab, "name",         PG_TEXT, 0, TRUE, FALSE,  1);
   add_column_to_table (tab, "descr",        PG_TEXT, 0, TRUE, FALSE,  1);
   add_column_to_table (tab, "install_date", PG_DATE, 0, TRUE, FALSE,  1);
   tables = g_list_append (tables, tab);

   /*
    * problem_type
    */
   tab = new_table_data ("problem_type", TRUE, 1);
   add_column_to_table (tab, "type_num",      PG_INT4,      0, FALSE, TRUE,  1);
   add_column_to_table (tab, "name",          PG_TEXT,      0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "severity_num",  PG_INT4,      0, FALSE, FALSE, 1);
   add_column_to_table (tab, "active",        PG_BOOL,      0, FALSE, FALSE, 1);
   add_column_to_table (tab, "descr",         PG_TEXT,      0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "creation_date", PG_TIMESTAMP, 0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "creator",       PG_NAME,      0, TRUE,  FALSE, 1);

   col = add_column_to_table (tab, "order_num", PG_INT4,
                              0, TRUE, FALSE, 4);
   col->prereq = FALSE;
   col->initsql = g_string_new ("");
   g_string_append_printf (col->initsql, INIT_ORDER_NUM, "problem_type",
                           "problem_type", "type_num", "problem_type",
                           "type_num");

   tables = g_list_append (tables, tab);

   /*
    * Person
    */
   tab = new_table_data ("person", TRUE, 1);
   add_column_to_table (tab, "login_id",      PG_NAME,      0, FALSE, TRUE,  1);
   add_column_to_table (tab, "last_name",     PG_TEXT,      0, FALSE, FALSE, 1);
   add_column_to_table (tab, "first_name",    PG_TEXT,      0, FALSE, FALSE, 1);
   add_column_to_table (tab, "middle_name",   PG_TEXT,      0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "e_mail",        PG_TEXT,      0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "active",        PG_BOOL,      0, FALSE, FALSE, 1);
   add_column_to_table (tab, "access_flag",   PG_INT4,      0, FALSE, FALSE, 1);
   add_column_to_table (tab, "creation_date", PG_TIMESTAMP, 0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "creator",       PG_NAME,      0, TRUE,  FALSE, 1);
   add_cons_to_table (tab,
         "person_access_flag",
         "CHECK (((access_flag >= 0) AND (access_flag <= 1)))");
   tables = g_list_append (tables, tab);

   /*
    * Project
    */
   tab = new_table_data ("project", TRUE, 1);
   add_column_to_table (tab, "project_num",   PG_INT4,      0, FALSE, TRUE,  1);
   add_column_to_table (tab, "name",          PG_TEXT,      0, FALSE, FALSE, 1);
   add_column_to_table (tab, "active",        PG_BOOL,      0, FALSE, FALSE, 1);
   add_column_to_table (tab, "login_id",      PG_NAME,      0, FALSE, FALSE, 1);
   add_column_to_table (tab, "descr",         PG_TEXT,      0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "creation_date", PG_TIMESTAMP, 0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "creator",       PG_NAME,      0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "new_pr_notify_grp",
                                              PG_BOOL,      0, TRUE,  FALSE, 2);
   add_column_to_table (tab, "new_pr_notify_pl",
                                              PG_BOOL,      0, TRUE,  FALSE, 2);
   add_column_to_table (tab, "assign_notify_resp",
                                              PG_BOOL,      0, TRUE,  FALSE, 2);
   add_column_to_table (tab, "assign_notify_sub",
                                              PG_BOOL,      0, TRUE,  FALSE, 2);
   add_column_to_table (tab, "assign_notify_pl",
                                              PG_BOOL,      0, TRUE,  FALSE, 2);
   add_column_to_table (tab, "upd_notify_grp",
                                              PG_BOOL,      0, TRUE,  FALSE, 2);
   add_column_to_table (tab, "upd_notify_pl", PG_BOOL,      0, TRUE,  FALSE, 2);
   add_column_to_table (tab, "upd_notify_resp",
                                              PG_BOOL,      0, TRUE,  FALSE, 2);
   add_column_to_table (tab, "upd_notify_sub",
                                              PG_BOOL,      0, TRUE,  FALSE, 2);
   add_column_to_table (tab, "open_version",  PG_INT4,      0, TRUE,  FALSE, 3);
   add_column_to_table (tab, "close_version", PG_INT4,      0, TRUE,  FALSE, 3);
   add_column_to_table (tab, "auto_assign_login_id",
                                              PG_NAME,      0, TRUE,  FALSE, 4);
   tables = g_list_append (tables, tab);

   /*
    * Problem Report
    */
   tab = new_table_data ("problem_report", TRUE, 1);
   add_column_to_table (tab, "problem_num",   PG_INT4,      0, FALSE, TRUE,  1);
   add_column_to_table (tab, "title",         PG_TEXT,      0, FALSE, FALSE, 1);
   add_column_to_table (tab, "descr",         PG_TEXT,      0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "fix",           PG_TEXT,      0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "project_num",   PG_INT4,      0, FALSE, FALSE, 1);
   add_column_to_table (tab, "severity_num",  PG_INT4,      0, FALSE, FALSE, 1);
   add_column_to_table (tab, "status_num",    PG_INT4,      0, FALSE, FALSE, 1);
   add_column_to_table (tab, "type_num",      PG_INT4,      0, FALSE, FALSE, 1);
   add_column_to_table (tab, "submitter_id",  PG_NAME,      0, FALSE, FALSE, 1);
   add_column_to_table (tab, "responsible_id",PG_NAME,      0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "locker_id",     PG_NAME,      0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "creation_date", PG_TIMESTAMP, 0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "creator",       PG_NAME,      0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "open_version",  PG_INT4,      0, TRUE,  FALSE, 3);
   add_column_to_table (tab, "close_version", PG_INT4,      0, TRUE,  FALSE, 3);
   add_column_to_table (tab, "change_log_text",
                                              PG_TEXT,      0, TRUE,  FALSE, 3);
   col = add_column_to_table (tab, "archived", PG_BOOL,     0, FALSE, FALSE, 5);
   col->prereq = FALSE;
   col->initsql = g_string_new
      ("update problem_report set archived = false where archived is NULL");

   tables = g_list_append (tables, tab);

   /*
    * Audit Trail
    */
   tab = new_table_data ("audit_trail", FALSE, 1);
   add_column_to_table (tab, "the_date",      PG_TIMESTAMP, 0, FALSE, TRUE,  1);
   add_column_to_table (tab, "problem_num",   PG_INT4,      0, FALSE, TRUE,  1);
   add_column_to_table (tab, "prev_state",    PG_INT4,      0, FALSE, FALSE, 1);
   add_column_to_table (tab, "new_state",     PG_INT4,      0, FALSE, FALSE, 1);
   add_column_to_table (tab, "prev_sevr",     PG_INT4,      0, FALSE, FALSE, 1);
   add_column_to_table (tab, "new_sevr",      PG_INT4,      0, FALSE, FALSE, 1);
   add_column_to_table (tab, "login_id",      PG_NAME,      0, FALSE, FALSE, 1);
   tables = g_list_append (tables, tab);

   /*
    * Submitter
    */
   tab = new_table_data ("submitter", FALSE, 1);
   add_column_to_table (tab, "project_num",   PG_INT4,      0, FALSE, TRUE,  1);
   add_column_to_table (tab, "login_id",      PG_NAME,      0, FALSE, TRUE,  1);
   add_column_to_table (tab, "creation_date", PG_TIMESTAMP, 0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "creator",       PG_NAME,      0, TRUE,  FALSE, 1);
   tables = g_list_append (tables, tab);

   /*
    * Responsible
    */
   tab = new_table_data ("responsible", FALSE, 1);
   add_column_to_table (tab, "project_num",   PG_INT4,      0, FALSE, TRUE,  1);
   add_column_to_table (tab, "login_id",      PG_NAME,      0, FALSE, TRUE,  1);
   add_column_to_table (tab, "creation_date", PG_TIMESTAMP, 0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "creator",       PG_NAME,      0, TRUE,  FALSE, 1);
   tables = g_list_append (tables, tab);

   /*
    * PR Query
    */
   tab = new_table_data ("pr_query", FALSE, 1);
   add_column_to_table (tab, "query_num",     PG_INT4,      0, FALSE, TRUE,  1);
   add_column_to_table (tab, "name",          PG_TEXT,      0, FALSE, FALSE, 1);
   add_column_to_table (tab, "descr",         PG_TEXT,      0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "is_system_query",
                                              PG_BOOL,      0, FALSE, FALSE, 1);
   add_column_to_table (tab, "is_raw_sql",    PG_BOOL,      0, FALSE, FALSE, 1);
   add_column_to_table (tab, "raw_sql",       PG_TEXT,      0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "severity_nums", PG_INT4_ARR,  0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "status_nums",   PG_INT4_ARR,  0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "problem_type_nums",
                                              PG_INT4_ARR,  0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "submitter_ids", PG_NAME_ARR,  0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "responsible_ids",
                                              PG_NAME_ARR,  0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "order_by",      PG_TEXT,      0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "creation_date", PG_TIMESTAMP, 0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "creator",       PG_NAME,      0, TRUE,  FALSE, 1);

   col = add_column_to_table (tab, "order_num", PG_INT4,
                              0, TRUE, FALSE, 4);
   col->prereq = FALSE;
   col->initsql = g_string_new ("");
   g_string_append_printf (col->initsql, INIT_ORDER_NUM, "pr_query",
                           "pr_query", "query_num", "pr_query",
                           "query_num");

   tables = g_list_append (tables, tab);

   /*
    * Project Version
    */
   tab = new_table_data ("project_version", FALSE, 3);
   add_column_to_table (tab, "version_num",   PG_INT4,      0, FALSE, TRUE,  3);
   add_column_to_table (tab, "project_num",   PG_INT4,      0, FALSE, TRUE,  3);
   add_column_to_table (tab, "version_text",  PG_CHRVAR,   32, TRUE,  FALSE, 3);
   add_column_to_table (tab, "descr",         PG_CHRVAR,  128, TRUE,  FALSE, 3);
   add_column_to_table (tab, "active",        PG_BOOL,      0, TRUE,  FALSE, 3);
   add_column_to_table (tab, "order_num",     PG_INT4,      0, TRUE,  FALSE, 3);
   add_column_to_table (tab, "creation_date", PG_TIMESTAMP, 0, TRUE,  FALSE, 3);
   add_column_to_table (tab, "creator",       PG_NAME,      0, TRUE,  FALSE, 3);
   tables = g_list_append (tables, tab);

   /*
    * Severity
    */
   tab = new_table_data ("severity", TRUE, 1);
   add_column_to_table (tab, "severity_num",  PG_INT4,      0, FALSE, TRUE,  1);
   add_column_to_table (tab, "name",          PG_TEXT,      0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "order_num",     PG_INT4,      0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "active",        PG_BOOL,      0, FALSE, FALSE, 1);
   add_column_to_table (tab, "descr",         PG_TEXT,      0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "creation_date", PG_TIMESTAMP, 0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "creator",       PG_NAME,      0, TRUE,  FALSE, 1);
   tables = g_list_append (tables, tab);

   /*
    * Status
    */
   tab = new_table_data ("status", TRUE, 1);
   add_column_to_table (tab, "status_num",    PG_INT4,      0, FALSE, TRUE,  1);
   add_column_to_table (tab, "name",          PG_TEXT,      0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "order_num",     PG_INT4,      0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "active",        PG_BOOL,      0, FALSE, FALSE, 1);
   add_column_to_table (tab, "descr",         PG_TEXT,      0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "creation_date", PG_TIMESTAMP, 0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "creator",       PG_NAME,      0, TRUE,  FALSE, 1);
   add_column_to_table (tab, "is_closed_status",
                                              PG_BOOL,      0, FALSE, FALSE, 3);
   tables = g_list_append (tables, tab);
}

static void
free_constraint (gpointer data, gpointer user_data)
{
   ConstraintData *cons = data;

   g_string_free (cons->name, TRUE);
   g_string_free (cons->condition, TRUE);
   g_free (cons);
}

static void
free_column (gpointer data, gpointer user_data)
{
   ColumnData *col = data;

   g_string_free (col->name, TRUE);
   if (col->initsql != NULL)
      {
      g_string_free (col->initsql, TRUE);
      }
   g_free (col);
}

static void
free_table (gpointer data, gpointer user_data)
{
   TableData *table = data;

   g_string_free (table->name, TRUE);
   g_list_foreach (table->columns, free_column, NULL);
   g_list_foreach (table->constraints, free_constraint, NULL);
   g_free (table);
}

static void
destroy_tables (void)
{
   g_list_foreach (tables, free_table, NULL);
   g_list_free (tables);
   tables = NULL;
}

gint
count_tables (void)
{
   return count_list_elements (tables);
}

TableData *
get_table (gint n)
{
   return g_list_nth_data (tables, n);
}

gint
count_table_cols (const TableData *tab)
{
   return count_list_elements (tab->columns);
}

ColumnData *
get_table_col (const TableData *tab, gint n)
{
   return g_list_nth_data (tab->columns, n);
}

/*
 * Language related functions
 */
static void
initialize_languages (void)
{
   LanguageData *lang;

   g_assert (languages == NULL);

   lang = g_malloc (sizeof (LanguageData));
   lang->name = g_string_new ("plpgsql");
   lang->handler = g_string_new ("plpgsql_call_handler");
   lang->compiler = g_string_new ("PL/pgSQL");
   lang->minver = 1;
   lang->maxver = NOVER;
   languages = g_list_append (languages, lang);
}


static void
free_language (gpointer data, gpointer user_data)
{
   LanguageData *lang = data;

   g_string_free (lang->name, TRUE);
   g_string_free (lang->handler, TRUE);
   g_string_free (lang->compiler, TRUE);
   g_free (lang);
}

static void
destroy_languages (void)
{
   g_list_foreach (languages, free_language, NULL);
   g_list_free (languages);
   languages = NULL;
}

gint
count_languages (void)
{
   return count_list_elements (languages);
}

LanguageData *
get_language (gint n)
{
   return g_list_nth_data (languages, n);
}



/*
 * Function related functions
 */
static void
initialize_functions (const gchar *plpgsql_path,
                      const gchar *libpreps_path)
{
   FunctionData *func;

   g_assert (functions == NULL);

   func = g_malloc (sizeof (FunctionData));
   func->name = g_string_new ("plpgsql_call_handler");
   func->args = g_string_new ("");
   func->rtn = g_string_new ("language_handler");
   func->libpath = g_string_new (plpgsql_path);
   func->lang = g_string_new ("C");
   func->minver = 1;
   func->maxver = NOVER;
   functions = g_list_append (functions, func);

   func = g_malloc (sizeof (FunctionData));
   func->name = g_string_new ("send_new_pr_msg");
   func->args = g_string_new ("text,text,text,text,text,text,text");
   func->rtn = g_string_new ("int4");
   func->libpath = g_string_new (libpreps_path);
   func->lang = g_string_new ("C");
   func->minver = 2;
   func->maxver = NOVER;
   functions = g_list_append (functions, func);

   func = g_malloc (sizeof (FunctionData));
   func->name = g_string_new ("send_assign_msg");
   func->args = g_string_new ("text,text,text,text,text,text,text,text");
   func->rtn = g_string_new ("int4");
   func->libpath = g_string_new (libpreps_path);
   func->lang = g_string_new ("C");
   func->minver = 2;
   func->maxver = NOVER;
   functions = g_list_append (functions, func);

   func = g_malloc (sizeof (FunctionData));
   func->name = g_string_new ("send_update_msg");
   func->args = g_string_new ("text,text,text,text,text,text,text,text");
   func->rtn = g_string_new ("int4");
   func->libpath = g_string_new (libpreps_path);
   func->lang = g_string_new ("C");
   func->minver = 2;
   func->maxver = NOVER;
   functions = g_list_append (functions, func);
}

static void
free_function (gpointer data, gpointer user_data)
{
   FunctionData *func = data;

   g_string_free (func->name, TRUE);
   g_string_free (func->args, TRUE);
   g_string_free (func->rtn, TRUE);
   g_string_free (func->libpath, TRUE);
   g_string_free (func->lang, TRUE);
   g_free (func);
}

static void
destroy_functions (void)
{
   g_list_foreach (functions, free_function, NULL);
   g_list_free (functions);
   functions = NULL;
}

gint
count_functions (void)
{
   return count_list_elements (functions);
}

FunctionData *
get_function (gint n)
{
   return g_list_nth_data (functions, n);
}


/*
 * Stored Function Related Functions
 */
static void
initialize_stored_funcs (void)
{
   StoredFuncData *sf;

   g_assert (stored_funcs == NULL);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("person_ri");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (PERSON_RI);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 1;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("problem_report_ri");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (PROBLEM_REPORT_RI);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 1;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("problem_type_ri");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (PROBLEM_TYPE_RI);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 1;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("project_ri");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (PROJECT_RI);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 1;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("severity_ri");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (SEVERITY_RI);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 1;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("status_ri");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (STATUS_RI);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 1;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("au_tr_status_ri");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (AU_TR_STATUS_RI);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 1;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("au_tr_sevr_ri");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (AU_TR_SEVR_RI);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 1;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("check_submitter");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (CHECK_SUBMITTER);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 1;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("check_responsible_ins");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (CHECK_RESPONSIBLE_INS);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 1;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("check_responsible_upd");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (CHECK_RESPONSIBLE_UPD);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 1;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("check_access");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (CHECK_ACCESS);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 1;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("check_access_del");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (CHECK_ACCESS_DEL);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 1;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("check_query_access");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (CHECK_QUERY_ACCESS);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 1;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("check_query_access_del");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (CHECK_QUERY_ACCESS_DEL);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 1;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("check_last_admin");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (CHECK_LAST_ADMIN);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 1;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("person_del_cleanup");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (PERSON_DEL_CLEANUP);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 1;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("pr_insert_active");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (PR_INSERT_ACTIVE);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 1;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("pr_update_active");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (PR_UPDATE_ACTIVE);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 1;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("creation_stamp");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (CREATION_STAMP);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 1;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("submitter_no_change");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (SUBMITTER_NO_CHANGE);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 1;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("check_severity_not_used");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (CHECK_SEVERITY_NOT_USED);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 1;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("check_status_not_used");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (CHECK_STATUS_NOT_USED);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 1;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("check_type_not_used");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (CHECK_TYPE_NOT_USED);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 1;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("new_pr_email_list");
   sf->args = g_string_new ("int4");
   sf->rtn = g_string_new ("text");
   sf->code = g_string_new (NEW_PR_EMAIL_LIST);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 2;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("new_pr_send_notification");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (NEW_PR_SEND_NOTIFICATION);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 2;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("assign_pr_email_list");
   sf->args = g_string_new ("int4, name, name");
   sf->rtn = g_string_new ("text");
   sf->code = g_string_new (ASSIGN_PR_EMAIL_LIST);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 2;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("update_pr_email_list");
   sf->args = g_string_new ("int4, name, name");
   sf->rtn = g_string_new ("text");
   sf->code = g_string_new (UPDATE_PR_EMAIL_LIST);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 2;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("prj_version_ri");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (PRJ_VERSION_RI);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 3;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("check_close_version");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (CHECK_CLOSE_VERSION);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 3;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("check_prj_version_not_used");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (CHECK_PRJ_VERSION_NOT_USED);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 3;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("next_version_num");
   sf->args = g_string_new ("int4");
   sf->rtn = g_string_new ("int4");
   sf->code = g_string_new (NEXT_VERSION_NUM);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 3;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("clean_up_project");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (CLEAN_UP_PROJECT);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 1;
   sf->maxver = 2;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("clean_up_project");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (CLEAN_UP_PROJECT_V3);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 3;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("auto_assign_pr");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (AUTO_ASSIGN_PR);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 4;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("auto_assign_ri");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (AUTO_ASSIGN_RI);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 4;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("check_query_name");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (CHECK_QUERY_NAME);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 4;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("can_edit_project");
   sf->args = g_string_new ("int4");
   sf->rtn = g_string_new ("bool");
   sf->code = g_string_new (CAN_EDIT_PROJECT);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 4;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("check_prj_access");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (CHECK_PRJ_ACCESS);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 1;
   sf->maxver = 3;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("check_prj_access");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (CHECK_PRJ_ACCESS_V4);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 4;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("assign_send_notification");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (ASSIGN_SEND_NOTIFICATION);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 2;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);

   sf = g_malloc (sizeof (StoredFuncData));
   sf->name = g_string_new ("update_send_notification");
   sf->args = g_string_new ("");
   sf->rtn = g_string_new ("trigger");
   sf->code = g_string_new (UPDATE_SEND_NOTIFICATION);
   sf->lang = g_string_new ("plpgsql");
   sf->minver = 2;
   sf->maxver = NOVER;
   stored_funcs = g_list_append (stored_funcs, sf);
}

static void
free_stored_func (gpointer data, gpointer user_data)
{
   StoredFuncData *sf = data;

   g_string_free (sf->name, TRUE);
   g_string_free (sf->args, TRUE);
   g_string_free (sf->rtn, TRUE);
   g_string_free (sf->code, TRUE);
   g_string_free (sf->lang, TRUE);
   g_free (sf);
}

static void
destroy_stored_funcs (void)
{
   g_list_foreach (stored_funcs, free_stored_func, NULL);
   g_list_free (stored_funcs);
   stored_funcs = NULL;
}

gint
count_stored_funcs (void)
{
   return count_list_elements (stored_funcs);
}

StoredFuncData *
get_stored_func (gint n)
{
   return g_list_nth_data (stored_funcs, n);
}


/*
 * Unique Index Related Functions
 */
static void
initialize_unique_indexes (void)
{
   UniqueIndexData *ui;

   g_assert (unique_indexes == NULL);

   ui = g_malloc (sizeof (UniqueIndexData));
   ui->name = g_string_new ("severity_name_key");
   ui->table = g_string_new ("severity");
   ui->column = g_string_new ("name");
   ui->minver = 1;
   ui->maxver = NOVER;
   unique_indexes = g_list_append (unique_indexes, ui);

   ui = g_malloc (sizeof (UniqueIndexData));
   ui->name = g_string_new ("status_name_key");
   ui->table = g_string_new ("status");
   ui->column = g_string_new ("name");
   ui->minver = 1;
   ui->maxver = NOVER;
   unique_indexes = g_list_append (unique_indexes, ui);

   ui = g_malloc (sizeof (UniqueIndexData));
   ui->name = g_string_new ("problem_type_name_key");
   ui->table = g_string_new ("problem_type");
   ui->column = g_string_new ("name");
   ui->minver = 1;
   ui->maxver = NOVER;
   unique_indexes = g_list_append (unique_indexes, ui);

   ui = g_malloc (sizeof (UniqueIndexData));
   ui->name = g_string_new ("project_name_key");
   ui->table = g_string_new ("project");
   ui->column = g_string_new ("name");
   ui->minver = 1;
   ui->maxver = NOVER;
   unique_indexes = g_list_append (unique_indexes, ui);
}

static void
free_unique_index (gpointer data, gpointer user_data)
{
   UniqueIndexData *ui = data;

   g_string_free (ui->name, TRUE);
   g_string_free (ui->table, TRUE);
   g_string_free (ui->column, TRUE);
   g_free (ui);
}

static void
destroy_unique_indexes (void)
{
   g_list_foreach (unique_indexes, free_unique_index, NULL);
   g_list_free (unique_indexes);
   unique_indexes = NULL;
}

gint
count_unique_indexes (void)
{
   return count_list_elements (unique_indexes);
}

UniqueIndexData *
get_unique_index (gint n)
{
   return g_list_nth_data (unique_indexes, n);
}


/*
 * Trigger Related Functions
 */
static void
initialize_triggers (void)
{
   TriggerData *td;

   g_assert (triggers == NULL);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("type_access_check_del");
   td->before = TRUE;
   td->ins = FALSE;
   td->upd = FALSE;
   td->del = TRUE;
   td->table = g_string_new ("problem_type");
   td->row = TRUE;
   td->func = g_string_new ("check_access_del");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("type_def_sevr_check");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("problem_type");
   td->row = TRUE;
   td->func = g_string_new ("severity_ri");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("stamp_type");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = FALSE;
   td->del = FALSE;
   td->table = g_string_new ("problem_type");
   td->row = TRUE;
   td->func = g_string_new ("creation_stamp");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("type_not_used_on_del");
   td->before = TRUE;
   td->ins = FALSE;
   td->upd = FALSE;
   td->del = TRUE;
   td->table = g_string_new ("problem_type");
   td->row = TRUE;
   td->func = g_string_new ("check_type_not_used");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("type_access_check");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("problem_type");
   td->row = TRUE;
   td->func = g_string_new ("check_access");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("stamp_person");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = FALSE;
   td->del = FALSE;
   td->table = g_string_new ("person");
   td->row = TRUE;
   td->func = g_string_new ("creation_stamp");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("person_check_last_admin");
   td->before = TRUE;
   td->ins = FALSE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("person");
   td->row = TRUE;
   td->func = g_string_new ("check_last_admin");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("person_del_cleanup");
   td->before = TRUE;
   td->ins = FALSE;
   td->upd = FALSE;
   td->del = TRUE;
   td->table = g_string_new ("person");
   td->row = TRUE;
   td->func = g_string_new ("person_del_cleanup");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("person_access_check");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("person");
   td->row = TRUE;
   td->func = g_string_new ("check_access");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("prj_access_check_ins");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = FALSE;
   td->del = FALSE;
   td->table = g_string_new ("project");
   td->row = TRUE;
   td->func = g_string_new ("check_access");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("project_leader_check");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("project");
   td->row = TRUE;
   td->func = g_string_new ("person_ri");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("stamp_project");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = FALSE;
   td->del = FALSE;
   td->table = g_string_new ("project");
   td->row = TRUE;
   td->func = g_string_new ("creation_stamp");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("prj_project_version_check");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("project");
   td->row = TRUE;
   td->func = g_string_new ("prj_version_ri");
   td->args = g_string_new ("");
   td->minver = 3;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("prj_access_check_del");
   td->before = TRUE;
   td->ins = FALSE;
   td->upd = FALSE;
   td->del = TRUE;
   td->table = g_string_new ("project");
   td->row = TRUE;
   td->func = g_string_new ("clean_up_project");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("auto_assign_id_check");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("project");
   td->row = TRUE;
   td->func = g_string_new ("auto_assign_ri");
   td->args = g_string_new ("");
   td->minver = 4;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("prj_access_check_update");
   td->before = TRUE;
   td->ins = FALSE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("project");
   td->row = TRUE;
   td->func = g_string_new ("check_prj_access");
   td->args = g_string_new ("");
   td->minver = 4;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("pr_sevr_check");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("problem_report");
   td->row = TRUE;
   td->func = g_string_new ("severity_ri");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("pr_type_check");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("problem_report");
   td->row = TRUE;
   td->func = g_string_new ("problem_type_ri");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("pr_proj_check");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("problem_report");
   td->row = TRUE;
   td->func = g_string_new ("project_ri");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("pr_submitter_check");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = FALSE;
   td->del = FALSE;
   td->table = g_string_new ("problem_report");
   td->row = TRUE;
   td->func = g_string_new ("check_submitter");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("pr_responsible_check_ins");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = FALSE;
   td->del = FALSE;
   td->table = g_string_new ("problem_report");
   td->row = TRUE;
   td->func = g_string_new ("check_responsible_ins");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("pr_responsible_check_upd");
   td->before = TRUE;
   td->ins = FALSE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("problem_report");
   td->row = TRUE;
   td->func = g_string_new ("check_responsible_upd");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("pr_status_check");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("problem_report");
   td->row = TRUE;
   td->func = g_string_new ("status_ri");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("pr_insert_check");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = FALSE;
   td->del = FALSE;
   td->table = g_string_new ("problem_report");
   td->row = TRUE;
   td->func = g_string_new ("pr_insert_active");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("pr_update_check");
   td->before = TRUE;
   td->ins = FALSE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("problem_report");
   td->row = TRUE;
   td->func = g_string_new ("pr_update_active");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("stamp_pr");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = FALSE;
   td->del = FALSE;
   td->table = g_string_new ("problem_report");
   td->row = TRUE;
   td->func = g_string_new ("creation_stamp");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("pr_submitter_nc");
   td->before = TRUE;
   td->ins = FALSE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("problem_report");
   td->row = TRUE;
   td->func = g_string_new ("submitter_no_change");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("new_pr_event_notification");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = FALSE;
   td->del = FALSE;
   td->table = g_string_new ("problem_report");
   td->row = TRUE;
   td->func = g_string_new ("new_pr_send_notification");
   td->args = g_string_new ("");
   td->minver = 2;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("pr_project_version_check");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("problem_report");
   td->row = TRUE;
   td->func = g_string_new ("prj_version_ri");
   td->args = g_string_new ("");
   td->minver = 3;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("pr_close_version_check");
   td->before = TRUE;
   td->ins = FALSE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("problem_report");
   td->row = TRUE;
   td->func = g_string_new ("check_close_version");
   td->args = g_string_new ("");
   td->minver = 3;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("auto_assign_pr_check");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = FALSE;
   td->del = FALSE;
   td->table = g_string_new ("problem_report");
   td->row = TRUE;
   td->func = g_string_new ("auto_assign_pr");
   td->args = g_string_new ("");
   td->minver = 4;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("assign_pr_event_notification");
   td->before = TRUE;
   td->ins = FALSE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("problem_report");
   td->row = TRUE;
   td->func = g_string_new ("assign_send_notification");
   td->args = g_string_new ("");
   td->minver = 2;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("update_pr_event_notification");
   td->before = TRUE;
   td->ins = FALSE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("problem_report");
   td->row = TRUE;
   td->func = g_string_new ("update_send_notification");
   td->args = g_string_new ("");
   td->minver = 2;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("aud_tr_person_check");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("audit_trail");
   td->row = TRUE;
   td->func = g_string_new ("person_ri");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("aud_tr_prob_rpt_check");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("audit_trail");
   td->row = TRUE;
   td->func = g_string_new ("problem_report_ri");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("aud_tr_status_check");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("audit_trail");
   td->row = TRUE;
   td->func = g_string_new ("au_tr_status_ri");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("aud_tr_sevr_checks");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("audit_trail");
   td->row = TRUE;
   td->func = g_string_new ("au_tr_sevr_ri");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("submitter_project_check");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("submitter");
   td->row = TRUE;
   td->func = g_string_new ("project_ri");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("submitter_person_check");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("submitter");
   td->row = TRUE;
   td->func = g_string_new ("person_ri");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("stamp_submitter");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = FALSE;
   td->del = FALSE;
   td->table = g_string_new ("submitter");
   td->row = TRUE;
   td->func = g_string_new ("creation_stamp");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("resp_project_check");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("responsible");
   td->row = TRUE;
   td->func = g_string_new ("project_ri");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("resp_person_check");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("responsible");
   td->row = TRUE;
   td->func = g_string_new ("person_ri");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("stamp_responsible");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = FALSE;
   td->del = FALSE;
   td->table = g_string_new ("responsible");
   td->row = TRUE;
   td->func = g_string_new ("creation_stamp");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("stamp_pr_query");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = FALSE;
   td->del = FALSE;
   td->table = g_string_new ("pr_query");
   td->row = TRUE;
   td->func = g_string_new ("creation_stamp");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("pr_query_check");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("pr_query");
   td->row = TRUE;
   td->func = g_string_new ("check_query_access");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("query_name_check");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("pr_query");
   td->row = TRUE;
   td->func = g_string_new ("check_query_name");
   td->args = g_string_new ("");
   td->minver = 4;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("pr_query_check_del");
   td->before = TRUE;
   td->ins = FALSE;
   td->upd = FALSE;
   td->del = TRUE;
   td->table = g_string_new ("pr_query");
   td->row = TRUE;
   td->func = g_string_new ("check_query_access_del");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("prj_version_not_used_on_del");
   td->before = TRUE;
   td->ins = FALSE;
   td->upd = FALSE;
   td->del = TRUE;
   td->table = g_string_new ("project_version");
   td->row = TRUE;
   td->func = g_string_new ("check_prj_version_not_used");
   td->args = g_string_new ("");
   td->minver = 3;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("sevr_access_check");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("severity");
   td->row = TRUE;
   td->func = g_string_new ("check_access");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("sevr_access_check_del");
   td->before = TRUE;
   td->ins = FALSE;
   td->upd = FALSE;
   td->del = TRUE;
   td->table = g_string_new ("severity");
   td->row = TRUE;
   td->func = g_string_new ("check_access_del");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("sevr_not_used_on_del");
   td->before = TRUE;
   td->ins = FALSE;
   td->upd = FALSE;
   td->del = TRUE;
   td->table = g_string_new ("severity");
   td->row = TRUE;
   td->func = g_string_new ("check_severity_not_used");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("stamp_sevr");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = FALSE;
   td->del = FALSE;
   td->table = g_string_new ("severity");
   td->row = TRUE;
   td->func = g_string_new ("creation_stamp");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("status_access_check");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = TRUE;
   td->del = FALSE;
   td->table = g_string_new ("status");
   td->row = TRUE;
   td->func = g_string_new ("check_access");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("status_access_check_del");
   td->before = TRUE;
   td->ins = FALSE;
   td->upd = FALSE;
   td->del = TRUE;
   td->table = g_string_new ("status");
   td->row = TRUE;
   td->func = g_string_new ("check_access_del");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("stat_not_used_on_del");
   td->before = TRUE;
   td->ins = FALSE;
   td->upd = FALSE;
   td->del = TRUE;
   td->table = g_string_new ("status");
   td->row = TRUE;
   td->func = g_string_new ("check_status_not_used");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);

   td = g_malloc (sizeof (TriggerData));
   td->name = g_string_new ("stamp_status");
   td->before = TRUE;
   td->ins = TRUE;
   td->upd = FALSE;
   td->del = FALSE;
   td->table = g_string_new ("status");
   td->row = TRUE;
   td->func = g_string_new ("creation_stamp");
   td->args = g_string_new ("");
   td->minver = 1;
   td->maxver = NOVER;
   triggers = g_list_append (triggers, td);
}

static void
free_trigger (gpointer data, gpointer user_data)
{
   TriggerData *td = data;

   g_string_free (td->name, TRUE);
   g_string_free (td->table, TRUE);
   g_string_free (td->func, TRUE);
   g_string_free (td->args, TRUE);
   g_free (td);
}

static void
destroy_triggers (void)
{
   g_list_foreach (triggers, free_trigger, NULL);
   g_list_free (triggers);
   triggers = NULL;
}

gint
count_triggers (void)
{
   return count_list_elements (triggers);
}

TriggerData *
get_trigger (gint n)
{
   return g_list_nth_data (triggers, n);
}


/*
 * Rule Related Functions
 */
static void
initialize_rules (void)
{
   RuleData *rd;

   g_assert (rules == NULL);

   rd = g_malloc (sizeof (RuleData));
   rd->name = g_string_new ("log_problem_update");
   rd->event = UPDATE_EVENT;
   rd->table = g_string_new ("problem_report");
   rd->where = g_string_new (LOG_RULE_WHERE);
   rd->instead = FALSE;
   rd->action = g_string_new (LOG_RULE_ACTION);
   rd->minver = 1;
   rd->maxver = NOVER;
   rules = g_list_append (rules, rd);
}

static void
free_rule (gpointer data, gpointer user_data)
{
   RuleData *td = data;

   g_string_free (td->name, TRUE);
   g_string_free (td->table, TRUE);
   g_string_free (td->where, TRUE);
   g_string_free (td->action, TRUE);
   g_free (td);
}

static void
destroy_rules (void)
{
   g_list_foreach (rules, free_rule, NULL);
   g_list_free (rules);
   rules = NULL;
}

gint
count_rules (void)
{
   return count_list_elements (rules);
}

RuleData *
get_rule (gint n)
{
   return g_list_nth_data (rules, n);
}


void
initialize_db_schema (void)
{
   initialize_sequences();
   initialize_tables();
   initialize_languages();
   initialize_functions(PLPGSQL_LIB_PATH, PREPS_LIB_PATH);
   initialize_stored_funcs();
   initialize_unique_indexes();
   initialize_triggers();
   initialize_rules();
}

void
cleanup_db_schema (void)
{
   destroy_sequences();
   destroy_tables();
   destroy_languages();
   destroy_functions();
   destroy_stored_funcs();
   destroy_unique_indexes();
   destroy_triggers();
   destroy_rules();
}

void
reset_library_paths (const gchar *plpgsql_path,
                     const gchar *libpreps_path)
{
   destroy_functions();
   initialize_functions(plpgsql_path, libpreps_path);
}
