#ifndef __DB_UPDATE_H
#define __DB_UPDATE_H
/*-------------------------------------------------------------------------
 * Copyright (c) 2004 Kenneth W. Sodemann (stuffle@mac.com)
 *-------------------------------------------------------------------------
 * db_update
 *
 * Synopsis:
 *    Functions used to update a database.
 *    
 * $Id: db_update.h,v 1.6 2004/06/29 18:45:41 stuffle Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330 
 * Boston, MA  02111-1307  USA
 *-------------------------------------------------------------------------
 */
#include <libpq-fe.h>
#include <glib.h>

#include "db_login.h"

/*-------------------------------------------------------------------------
 * drop_X
 *
 * Drop all of the objects of type X in the database.
 *
 * Inputs:
 *    newdb_data : information about the database, must be connected
 *                 as the owner of the schema
 *    -OR-
 *    pgsu_data : information about the database, must be connected
 *                to the databas as the PostgreSQL super user
 *
 * Returns the result of the last statement executed.  This value be
 * examined and PQclear()'ed by the caller.
 *
 * In cases where there is nothing to drop, the function will return
 * NULL.  The caller should test for NULL results before examining or
 * PQclear()'ing the result.
 *
 *-------------------------------------------------------------------------
 */
PGresult *drop_indexes (DbLoginData *newdb_data);
PGresult *drop_primary_keys (DbLoginData *newdb_data);
PGresult *drop_rules (DbLoginData *newdb_data);
PGresult *drop_triggers (DbLoginData *newdb_data);

PGresult *drop_stored_procedures (DbLoginData *newdb_data);

PGresult *drop_external_functions (DbLoginData *pgsu_data);
PGresult *drop_languages (DbLoginData *pgsu_data);

/*-------------------------------------------------------------------------
 * add_missing_X
 *
 * Look for missing objects and add them. 
 *
 * Inputs:
 *    newdb_data : information about the database, must be connected
 *                 as the owner of the schema
 *    version    : database schema version
 *
 * Returns the result of the last statement executed.  This value be
 * examined and PQclear()'ed by the caller.
 *
 * In cases where there is nothing to add, the function will return
 * NULL.  The caller should test for NULL results before examining or
 * PQclear()'ing the result.
 *
 *-------------------------------------------------------------------------
 */
PGresult *add_missing_tables (DbLoginData *newdb_data, gint version);
PGresult *add_missing_cols (DbLoginData *newdb_data, gint version);
PGresult *add_missing_primary_keys (DbLoginData *newdb_data, gint version);
PGresult *add_missing_indexes (DbLoginData *newdb_data, gint version);
#endif
