/*-------------------------------------------------------------------------
 * Copyright (c) 2003-2004 Kenneth W. Sodemann (stuffle@mac.com)
 *-------------------------------------------------------------------------
 * progdlg
 *
 * Synopsis:
 *   The database creation progress dialog.
 *    
 * $Id: progdlg.c,v 1.8 2004/08/05 03:33:58 stuffle Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330 
 * Boston, MA  02111-1307  USA
 *-------------------------------------------------------------------------
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>
#include "gtkutils.h"
#include "progdlg.h"

#define CREATE_DB_LABEL "create_db_label"
#define CREATE_BASIC_DATA_LABEL "create_basic_data_label"
#define CREATE_ADMIN_USER_LABEL "create_admin_user_label"
#define INSTALL_LIBS_LABEL "install_libs_label"
#define INSTALL_PROC_LABEL "install_proc_label"
#define INSTALL_INIT_DATA_LABEL "install_init_data_label"

#define CREATE_DB_STAT_LABEL "create_db_stat_label"
#define CREATE_BASIC_DATA_STAT_LABEL "create_basic_data_stat_label"
#define CREATE_ADMIN_USER_STAT_LABEL "create_admin_user_stat_label"
#define INSTALL_LIBS_STAT_LABEL "install_libs_stat_label"
#define INSTALL_PROC_STAT_LABEL "install_proc_stat_label"
#define INSTALL_INIT_DATA_STAT_LABEL "install_init_data_stat_label"

#define UPDATE_INIT_LABEL "update_init_label"
#define DROP_OBJECTS_LABEL "drop_objects_label"
#define UPDATE_SCHEMA_LABEL "update_schema_label"
#define REINSTALL_LIBS_LABEL "reinstall_libs_label"
#define REINSTALL_PROCS_LABEL "reinstall_procs_label"
#define REINSTALL_TRIGS_LABEL "reinstall_trigs_label"
#define INIT_NEW_COLS_LABEL "init_new_cols_label"
#define UPDATE_VER_INFO_LABEL "update_ver_info_label"

#define UPDATE_INIT_STAT_LABEL "update_init_stat_label"
#define DROP_OBJECTS_STAT_LABEL "drop_objects_stat_label"
#define UPDATE_SCHEMA_STAT_LABEL "update_schema_stat_label"
#define REINSTALL_LIBS_STAT_LABEL "reinstall_libs_stat_label"
#define REINSTALL_PROCS_STAT_LABEL "reinstall_procs_stat_label"
#define REINSTALL_TRIGS_STAT_LABEL "reinstall_trigs_stat_label"
#define INIT_NEW_COLS_STAT_LABEL "init_new_cols_stat_label"
#define UPDATE_VER_INFO_STAT_LABEL "update_ver_info_stat_label"

#define CLOSE_BUTTON "close_button"


static void
on_close_btn_clicked (GtkButton *btn,
                      gpointer   user_data)
{
   GtkWidget *main_win;

   main_win = lookup_widget (GTK_WIDGET (user_data), MY_PARENT);
   g_assert (main_win != FALSE);

   gtk_widget_destroy (main_win);
}


static void
on_dlg_destroy (GObject *obj, gpointer user_data)
{
   GtkWidget *main_win;

   main_win = lookup_widget (GTK_WIDGET (obj), MY_PARENT);
   g_assert (main_win != FALSE);

   gtk_widget_destroy (main_win);
}


void
set_creation_step_status (GtkWidget    *dlg,
                          CreationStep  step,
                          StepStatus    status)
{
   GtkLabel    *stat_label = NULL;
   GtkLabel    *step_label = NULL;
   const gchar *text;
   GString     *str;

   switch (step)
      {
      case CREATE_DATABASE:
         stat_label = GTK_LABEL (lookup_widget (dlg, CREATE_DB_STAT_LABEL));
         step_label = GTK_LABEL (lookup_widget (dlg, CREATE_DB_LABEL));
         break;
      case CREATE_DB_STRUCTURE:
         stat_label =
            GTK_LABEL (lookup_widget (dlg, CREATE_BASIC_DATA_STAT_LABEL));
         step_label =
            GTK_LABEL (lookup_widget (dlg, CREATE_BASIC_DATA_LABEL));
         break;
      case CREATE_ADMIN_USER:
         stat_label =
            GTK_LABEL (lookup_widget (dlg, CREATE_ADMIN_USER_STAT_LABEL));
         step_label =
            GTK_LABEL (lookup_widget (dlg, CREATE_ADMIN_USER_LABEL));
         break;
      case INSTALL_LIBS:
         stat_label = GTK_LABEL (lookup_widget (dlg, INSTALL_LIBS_STAT_LABEL));
         step_label = GTK_LABEL (lookup_widget (dlg, INSTALL_LIBS_LABEL));
         break;
      case INSTALL_PROCS:
         stat_label = GTK_LABEL (lookup_widget (dlg, INSTALL_PROC_STAT_LABEL));
         step_label = GTK_LABEL (lookup_widget (dlg, INSTALL_PROC_LABEL));
         break;
      case INSTALL_INIT_DATA:
         stat_label =
            GTK_LABEL (lookup_widget (dlg, INSTALL_INIT_DATA_STAT_LABEL));
         step_label =
            GTK_LABEL (lookup_widget (dlg, INSTALL_INIT_DATA_LABEL));
         break;
      default: 
         g_assert_not_reached();
      }
   g_assert (stat_label != NULL);
   g_assert (step_label != NULL);

   text = gtk_label_get_text (step_label);
   str = g_string_new ("");

   switch (status)
      {
      case STEP_NO_STATUS:
         gtk_label_set_text (stat_label, "");
         g_string_printf (str, "%s", text);
         break;
      case STEP_RUNNING:
         gtk_label_set_markup (stat_label, _("<b>(Running)</b>"));
         g_string_printf (str, "<b>%s</b>", text);
         break;
      case STEP_SUCCESS:
         gtk_label_set_text (stat_label, _("(Success)"));
         g_string_printf (str, "%s", text);
         break;
      case STEP_FAILED:
         gtk_label_set_markup
            (stat_label,
             _("<span foreground=\"red\" weight=\"bold\">(Failed)</span>"));
         g_string_printf
            (str, "<span foreground=\"red\" weight=\"bold\">%s</span>", text);
         break;
      default:
         g_assert_not_reached();
      }

   gtk_label_set_markup (step_label, str->str);
   g_string_free (str, TRUE);
}


void
set_update_step_status (GtkWidget  *dlg,
                        UpdateStep  step,
                        StepStatus  status)
{
   GtkLabel    *stat_label = NULL;
   GtkLabel    *step_label = NULL;
   const gchar *text;
   GString     *str;

   switch (step)
      {
      case UPDATE_INIT:
         stat_label = GTK_LABEL (lookup_widget (dlg, UPDATE_INIT_STAT_LABEL));
         step_label = GTK_LABEL (lookup_widget (dlg, UPDATE_INIT_LABEL));
         break;
      case DROP_OBJECTS:
         stat_label = GTK_LABEL (lookup_widget (dlg, DROP_OBJECTS_STAT_LABEL));
         step_label = GTK_LABEL (lookup_widget (dlg, DROP_OBJECTS_LABEL));
         break;
      case UPDATE_SCHEMA:
         stat_label = GTK_LABEL (lookup_widget (dlg, UPDATE_SCHEMA_STAT_LABEL));
         step_label = GTK_LABEL (lookup_widget (dlg, UPDATE_SCHEMA_LABEL));
         break;
      case REINSTALL_LIBS:
         stat_label =
            GTK_LABEL (lookup_widget (dlg, REINSTALL_LIBS_STAT_LABEL));
         step_label =
            GTK_LABEL (lookup_widget (dlg, REINSTALL_LIBS_LABEL));
         break;
      case REINSTALL_PROCS:
         stat_label =
            GTK_LABEL (lookup_widget (dlg, REINSTALL_PROCS_STAT_LABEL));
         step_label =
            GTK_LABEL (lookup_widget (dlg, REINSTALL_PROCS_LABEL));
         break;
      case REINSTALL_TRIGS:
         stat_label =
            GTK_LABEL (lookup_widget (dlg, REINSTALL_TRIGS_STAT_LABEL));
         step_label =
            GTK_LABEL (lookup_widget (dlg, REINSTALL_TRIGS_LABEL));
         break;
      case UPDATE_VER_INFO:
         stat_label =
            GTK_LABEL (lookup_widget (dlg, UPDATE_VER_INFO_STAT_LABEL));
         step_label =
            GTK_LABEL (lookup_widget (dlg, UPDATE_VER_INFO_LABEL));
         break;
      default: 
         g_assert_not_reached();
      }
   g_assert (stat_label != NULL);
   g_assert (step_label != NULL);

   text = gtk_label_get_text (step_label);
   str = g_string_new ("");

   switch (status)
      {
      case STEP_NO_STATUS:
         gtk_label_set_text (stat_label, "");
         g_string_printf (str, "%s", text);
         break;
      case STEP_RUNNING:
         gtk_label_set_markup (stat_label, _("<b>(Running)</b>"));
         g_string_printf (str, "<b>%s</b>", text);
         break;
      case STEP_SUCCESS:
         gtk_label_set_text (stat_label, _("(Success)"));
         g_string_printf (str, "%s", text);
         break;
      case STEP_FAILED:
         gtk_label_set_markup
            (stat_label,
             _("<span foreground=\"red\" weight=\"bold\">(Failed)</span>"));
         g_string_printf
            (str, "<span foreground=\"red\" weight=\"bold\">%s</span>", text);
         break;
      default:
         g_assert_not_reached();
      }

   gtk_label_set_markup (step_label, str->str);
   g_string_free (str, TRUE);
}


void
activate_close_btn (GtkWidget *dlg)
{
   GtkWidget *button;

   button = lookup_widget (dlg, CLOSE_BUTTON);
   g_assert (button != NULL);

   gtk_widget_set_sensitive (button, TRUE);
}



   
GtkWidget *
create_prog_dlg (GtkWindow *parent, ProgressDlgType type)
{
   GtkWidget    *window;
   GtkWidget    *vbox;
   GtkWidget    *hbox;
   GtkWidget    *label;
   GtkWidget    *hseparator;
   GtkWidget    *hbuttonbox;
   GtkWidget    *button;
   GtkSizeGroup *size_group;

   window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
   gtk_container_set_border_width (GTK_CONTAINER (window), 12);
   gtk_window_set_title (GTK_WINDOW (window), _("Database Creation Status"));
   gtk_window_set_position (GTK_WINDOW (window), GTK_WIN_POS_CENTER);
   gtk_window_set_resizable (GTK_WINDOW (window), FALSE);
   gtk_window_set_transient_for (GTK_WINDOW (window), parent);
   gtk_window_set_modal (GTK_WINDOW (window), TRUE);
   GLADE_HOOKUP_OBJECT_NO_REF (window, window, TOP_DOG);
   GLADE_HOOKUP_OBJECT_NO_REF (window, parent, MY_PARENT);
   g_signal_connect (window, "destroy", G_CALLBACK (on_dlg_destroy), NULL);

   size_group = gtk_size_group_new (GTK_SIZE_GROUP_HORIZONTAL);

   vbox = gtk_vbox_new (FALSE, 6);
   gtk_widget_show (vbox);
   gtk_container_add (GTK_CONTAINER (window), vbox);
   GLADE_HOOKUP_OBJECT (window, vbox, "vbox1");

   if (type == CREATE_DLG)
      {
      hbox = gtk_hbox_new (FALSE, 3);
      gtk_widget_show (hbox);
      gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
      GLADE_HOOKUP_OBJECT (window, hbox, "cr8hbox1");

      label = gtk_label_new (_("Create Database"));
      gtk_widget_show (label);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      GLADE_HOOKUP_OBJECT (window, label, CREATE_DB_LABEL);
      gtk_size_group_add_widget (size_group, label);

      label = gtk_label_new ("");
      gtk_widget_show (label);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      GLADE_HOOKUP_OBJECT (window, label, CREATE_DB_STAT_LABEL);

      hbox = gtk_hbox_new (FALSE, 3);
      gtk_widget_show (hbox);
      gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
      GLADE_HOOKUP_OBJECT (window, hbox, "cr8hbox2");

      label = gtk_label_new (_("Create Basic Data Structures"));
      gtk_widget_show (label);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      GLADE_HOOKUP_OBJECT (window, label, CREATE_BASIC_DATA_LABEL);
      gtk_size_group_add_widget (size_group, label);

      label = gtk_label_new ("");
      gtk_widget_show (label);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      GLADE_HOOKUP_OBJECT (window, label, CREATE_BASIC_DATA_STAT_LABEL);

      hbox = gtk_hbox_new (FALSE, 3);
      gtk_widget_show (hbox);
      gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
      GLADE_HOOKUP_OBJECT (window, hbox, "cr8hbox3");

      label = gtk_label_new (_("Create Default Admin User"));
      gtk_widget_show (label);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      GLADE_HOOKUP_OBJECT (window, label, CREATE_ADMIN_USER_LABEL);
      gtk_size_group_add_widget (size_group, label);

      label = gtk_label_new ("");
      gtk_widget_show (label);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      GLADE_HOOKUP_OBJECT (window, label, CREATE_ADMIN_USER_STAT_LABEL);

      hbox = gtk_hbox_new (FALSE, 3);
      gtk_widget_show (hbox);
      gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
      GLADE_HOOKUP_OBJECT (window, hbox, "cr8hbox4");

      label = gtk_label_new (_("Install Procedural Languages and Libraries"));
      gtk_widget_show (label);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      GLADE_HOOKUP_OBJECT (window, label, INSTALL_LIBS_LABEL);
      gtk_size_group_add_widget (size_group, label);

      label = gtk_label_new ("");
      gtk_widget_show (label);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      GLADE_HOOKUP_OBJECT (window, label, INSTALL_LIBS_STAT_LABEL);

      hbox = gtk_hbox_new (FALSE, 3);
      gtk_widget_show (hbox);
      gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
      GLADE_HOOKUP_OBJECT (window, hbox, "cr8hbox5");

      label = gtk_label_new (_("Install Stored Procedures"));
      gtk_widget_show (label);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      GLADE_HOOKUP_OBJECT (window, label, INSTALL_PROC_LABEL);
      gtk_size_group_add_widget (size_group, label);

      label = gtk_label_new ("");
      gtk_widget_show (label);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      GLADE_HOOKUP_OBJECT (window, label, INSTALL_PROC_STAT_LABEL);

      hbox = gtk_hbox_new (FALSE, 3);
      gtk_widget_show (hbox);
      gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
      GLADE_HOOKUP_OBJECT (window, hbox, "cr8hbox6");

      label = gtk_label_new (_("Install Initial Data"));
      gtk_widget_show (label);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      GLADE_HOOKUP_OBJECT (window, label, INSTALL_INIT_DATA_LABEL);
      gtk_size_group_add_widget (size_group, label);

      label = gtk_label_new ("");
      gtk_widget_show (label);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      GLADE_HOOKUP_OBJECT (window, label, INSTALL_INIT_DATA_STAT_LABEL);
      }
   else
      {
      hbox = gtk_hbox_new (FALSE, 3);
      gtk_widget_show (hbox);
      gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
      GLADE_HOOKUP_OBJECT (window, hbox, "updhbox1");

      label = gtk_label_new (_("Initialization"));
      gtk_widget_show (label);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      GLADE_HOOKUP_OBJECT (window, label, UPDATE_INIT_LABEL);
      gtk_size_group_add_widget (size_group, label);

      label = gtk_label_new ("");
      gtk_widget_show (label);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      GLADE_HOOKUP_OBJECT (window, label, UPDATE_INIT_STAT_LABEL);

      hbox = gtk_hbox_new (FALSE, 3);
      gtk_widget_show (hbox);
      gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
      GLADE_HOOKUP_OBJECT (window, hbox, "updhbox2");

      label = gtk_label_new (_("Drop Objects"));
      gtk_widget_show (label);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      GLADE_HOOKUP_OBJECT (window, label, DROP_OBJECTS_LABEL);
      gtk_size_group_add_widget (size_group, label);

      label = gtk_label_new ("");
      gtk_widget_show (label);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      GLADE_HOOKUP_OBJECT (window, label, DROP_OBJECTS_STAT_LABEL);

      hbox = gtk_hbox_new (FALSE, 3);
      gtk_widget_show (hbox);
      gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
      GLADE_HOOKUP_OBJECT (window, hbox, "updhbox3");

      label = gtk_label_new (_("Update Schema"));
      gtk_widget_show (label);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      GLADE_HOOKUP_OBJECT (window, label, UPDATE_SCHEMA_LABEL);
      gtk_size_group_add_widget (size_group, label);

      label = gtk_label_new ("");
      gtk_widget_show (label);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      GLADE_HOOKUP_OBJECT (window, label, UPDATE_SCHEMA_STAT_LABEL);

      hbox = gtk_hbox_new (FALSE, 3);
      gtk_widget_show (hbox);
      gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
      GLADE_HOOKUP_OBJECT (window, hbox, "updhbox4");

      label = gtk_label_new (_("Reinstall Libraries"));
      gtk_widget_show (label);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      GLADE_HOOKUP_OBJECT (window, label, REINSTALL_LIBS_LABEL);
      gtk_size_group_add_widget (size_group, label);

      label = gtk_label_new ("");
      gtk_widget_show (label);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      GLADE_HOOKUP_OBJECT (window, label, REINSTALL_LIBS_STAT_LABEL);

      hbox = gtk_hbox_new (FALSE, 3);
      gtk_widget_show (hbox);
      gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
      GLADE_HOOKUP_OBJECT (window, hbox, "updhbox5");

      label = gtk_label_new (_("Reinstall Functions"));
      gtk_widget_show (label);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      GLADE_HOOKUP_OBJECT (window, label, REINSTALL_PROCS_LABEL);
      gtk_size_group_add_widget (size_group, label);

      label = gtk_label_new ("");
      gtk_widget_show (label);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      GLADE_HOOKUP_OBJECT (window, label, REINSTALL_PROCS_STAT_LABEL);

      hbox = gtk_hbox_new (FALSE, 3);
      gtk_widget_show (hbox);
      gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
      GLADE_HOOKUP_OBJECT (window, hbox, "updhbox6");

      label = gtk_label_new (_("Reinstall Triggers"));
      gtk_widget_show (label);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      GLADE_HOOKUP_OBJECT (window, label, REINSTALL_TRIGS_LABEL);
      gtk_size_group_add_widget (size_group, label);

      label = gtk_label_new ("");
      gtk_widget_show (label);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      GLADE_HOOKUP_OBJECT (window, label, REINSTALL_TRIGS_STAT_LABEL);

      hbox = gtk_hbox_new (FALSE, 3);
      gtk_widget_show (hbox);
      gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
      GLADE_HOOKUP_OBJECT (window, hbox, "updhbox7");

      label = gtk_label_new (_("Update Version Information"));
      gtk_widget_show (label);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      GLADE_HOOKUP_OBJECT (window, label, UPDATE_VER_INFO_LABEL);
      gtk_size_group_add_widget (size_group, label);

      label = gtk_label_new ("");
      gtk_widget_show (label);
      gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
      GLADE_HOOKUP_OBJECT (window, label, UPDATE_VER_INFO_STAT_LABEL);
#if 0
      hbox = gtk_hbox_new (FALSE, 3);
      gtk_widget_show (hbox);
      gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
      GLADE_HOOKUP_OBJECT (window, hbox, "updhbox8");
#endif
      }

   hseparator = gtk_hseparator_new ();
   gtk_widget_show (hseparator);
   gtk_box_pack_start (GTK_BOX (vbox), hseparator, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, hseparator, "hseparator");

   hbuttonbox = gtk_hbutton_box_new ();
   gtk_widget_show (hbuttonbox);
   gtk_box_pack_start (GTK_BOX (vbox), hbuttonbox, FALSE, FALSE, 0);
   GLADE_HOOKUP_OBJECT (window, hbuttonbox, "hbuttonbox");

   button = gtk_button_new_from_stock ("gtk-close");
   gtk_widget_show (button);
   gtk_container_add (GTK_CONTAINER (hbuttonbox), button);
   gtk_widget_set_sensitive (button, FALSE);
   GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
   GLADE_HOOKUP_OBJECT (window, button, CLOSE_BUTTON);
   g_signal_connect (GTK_OBJECT (button), "clicked",
                     G_CALLBACK (on_close_btn_clicked), window);

   return window;
}
